from tornado.httputil import parse_cookie
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application, RequestHandler

from pcs_test.tier0.daemon.test_session import AssertMixin
from pcs_test.tier0.daemon.app.fixtures_app import(
    USER,
    GROUPS,
    PASSWORD,
    UserAuthInfo,
    UserAuthMixin
)
from pcs_test.tools.misc import create_setup_patch_mixin

from pcs.daemon import session
from pcs.daemon.app import session as app_session

# pylint: disable=too-many-ancestors

SID = "abc"

def headers_with_sid(cookie_sid):
    return {
        "headers": {
            "Cookie": f"{app_session.PCSD_SESSION}={cookie_sid}"
        }
    }

class Handler(app_session.Mixin, RequestHandler):
    # pylint: disable=abstract-method,arguments-differ
    test = None
    @property
    def response_cookies(self):
        return {} if not hasattr(self, "_new_cookie") else {
            key: morsel.value for key, morsel in self._new_cookie.items()
        }

    async def get(self, *args, **kwargs):
        del args, kwargs
        if self.test.auto_init_session:
            await self.init_session()
        await self.test.on_handle(self)

VANILLA_SESSION = "VANILLA_SESSION"
AUTHENTICATED_SESSION = "AUTHENTICATED_SESSION"
RESPONSE_WITH_SID = "RESPONSE_WITH_SID"
RESPONSE_WITHOUT_SID = "RESPONSE_WITHOUT_SID"
RESPONSE_SID_IN_STORAGE = "RESPONSE_SID_IN_STORAGE"

class MixinTest(
    AsyncHTTPTestCase, AssertMixin, create_setup_patch_mixin(app_session),
    UserAuthMixin
):
    init_session = None
    auto_init_session = True
    groups_valid = False
    response_sid_expectation = None
    """
    The app_session.Mixin is tested via Handler(RequestHandler) that mix it. The
    particular tests acts inside the handler.
    """
    def setUp(self):
        Handler.test = self
        self.storage = session.Storage(lifetime_seconds=10)
        self.setup_patch("check_user_groups", self.check_user_groups)
        self.setup_patch("authorize_user", self.authorize_user)
        self.fetch_args = {}
        self.sid = None

        self.response = None

        super().setUp()

    def get_app(self):
        return Application([("/", Handler, dict(session_storage=self.storage))])

    @property
    def session_dict(self):
        # pylint: disable=protected-access
        return self.storage._Storage__sessions

    def sid_from_body(self, response):
        self.assertIn("Set-Cookie", response.headers)
        cookie = parse_cookie(response.headers["Set-Cookie"])
        self.assertIn(app_session.PCSD_SESSION, cookie)
        return cookie[app_session.PCSD_SESSION]

    def extra_checks(self, response):
        pass

    async def on_handle(self, handler):
        pass

    def test(self):
        if self.init_session == VANILLA_SESSION:
            self.sid = self.storage.provide().sid
            self.fetch_args = headers_with_sid(self.sid)
        elif self.init_session == AUTHENTICATED_SESSION:
            self.sid = self.storage.login(
                sid=None,
                username=USER,
                groups=[]
            ).sid
            self.fetch_args = headers_with_sid(self.sid)

        response = self.fetch("/", **self.fetch_args)
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b"")
        if self.response_sid_expectation == RESPONSE_WITH_SID:
            self.assertEqual(self.sid_from_body(response), self.sid)
        elif self.response_sid_expectation == RESPONSE_WITHOUT_SID:
            self.assertNotIn("Set-Cookie", response.headers)
        self.extra_checks(response)

class SidNotInRequestCookiesByDefault(MixinTest):
    response_sid_expectation = RESPONSE_WITHOUT_SID
    async def on_handle(self, handler):
        self.assertFalse(handler.was_sid_in_request_cookies())

class SidInRequestCookiesButNotInResponseCookies(MixinTest):
    init_session = VANILLA_SESSION
    response_sid_expectation = RESPONSE_WITHOUT_SID
    async def on_handle(self, handler):
        self.assertTrue(handler.was_sid_in_request_cookies())
        self.assertEqual(self.sid, handler.session.sid)
        self.assertFalse(app_session.PCSD_SESSION in handler.response_cookies)

class SidInResponseCookies(MixinTest):
    init_session = VANILLA_SESSION
    response_sid_expectation = RESPONSE_WITH_SID
    async def on_handle(self, handler):
        self.assertTrue(handler.was_sid_in_request_cookies())
        handler.put_request_cookies_sid_to_response_cookies_sid()
        self.assertTrue(app_session.PCSD_SESSION in handler.response_cookies)

class GetNewSession(MixinTest):
    auto_init_session = False
    response_sid_expectation = RESPONSE_WITH_SID
    async def on_handle(self, handler):
        self.assertEqual(len(self.session_dict), 0)
        await handler.init_session()
        self.assertEqual(len(self.session_dict), 1)
        self.sid = handler.session.sid
        handler.sid_to_cookies()

class SessionInMixinSurviveDestroyInStorage(MixinTest):
    async def on_handle(self, handler):
        # A handler can wait for something (e.g. external command). The already
        # started request should succeed even if the session expires in the
        # meantime.
        session1 = handler.session
        self.storage.destroy(session1.sid)
        self.assertIs(session1, handler.session)

class SessionIsPropagatedToResponseCookie(MixinTest):
    async def on_handle(self, handler):
        handler.sid_to_cookies()
        self.assertEqual(
            handler.response_cookies[app_session.PCSD_SESSION],
            handler.session.sid
        )

    def extra_checks(self, response):
        self.assertIn(self.sid_from_body(response), self.session_dict)

class CanLoginAndLogout(MixinTest):
    user_auth_info = UserAuthInfo(valid=True)
    auto_init_session = False
    async def on_handle(self, handler):
        await handler.session_auth_user(USER, PASSWORD)
        self.assert_authenticated_session(handler.session, USER, GROUPS)
        handler.session_logout()
        self.assert_vanila_session(handler.session)

class FailedLoginAttempt(MixinTest):
    auto_init_session = False
    async def on_handle(self, handler):
        await handler.session_auth_user(USER, PASSWORD)
        self.assert_login_failed_session(handler.session, USER)

class FailedLoginAttemptWithoutSessionSign(MixinTest):
    auto_init_session = False
    async def on_handle(self, handler):
        await handler.session_auth_user(
            USER,
            PASSWORD,
            sign_rejection=False,
        )
        self.assert_vanila_session(handler.session)

class CanLogoutWithoutSessionAccess(MixinTest):
    init_session = VANILLA_SESSION
    auto_init_session = False
    async def on_handle(self, handler):
        handler.session_logout()
        self.assertNotEqual(self.sid, handler.session.sid)

class AuthUpdatedByGroupCheck(MixinTest):
    init_session = AUTHENTICATED_SESSION
    groups_valid = True

    async def on_handle(self, handler):
        self.assertTrue(handler.was_sid_in_request_cookies())
        self.assertTrue(handler.session.is_authenticated)

class AuthRefusedByGroupCheck(MixinTest):
    init_session = AUTHENTICATED_SESSION
    async def on_handle(self, handler):
        self.assertFalse(handler.session.is_authenticated)



from pcs.cli.constraint.console_report import (
    constraint_plain as constraint_plain_default,
    constraint_with_sets,
)
from pcs.cli.constraint_colocation.console_report import (
    constraint_plain as colocation_plain
)
from pcs.cli.constraint_order.console_report import (
    constraint_plain as order_plain
)
from pcs.cli.constraint_ticket.console_report import (
    constraint_plain as ticket_plain
)
from pcs.common import report_codes as codes


def constraint(constraint_type, constraint_info, with_id=True):
    """
    dict constraint_info  see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    if "resource_sets" in constraint_info:
        return constraint_with_sets(constraint_info, with_id)
    return constraint_plain(constraint_type, constraint_info, with_id)

def constraint_plain(constraint_type, options_dict, with_id=False):
    """return console shape for any constraint_type of plain constraint"""
    type_report_map = {
        "rsc_colocation": colocation_plain,
        "rsc_order": order_plain,
        "rsc_ticket": ticket_plain,
    }

    if constraint_type not in type_report_map:
        return constraint_plain_default(constraint_type, options_dict, with_id)

    return type_report_map[constraint_type](options_dict, with_id)

#Each value (a callable taking report_item.info) returns a message.
#Force text will be appended if necessary.
#If it is necessary to put the force text inside the string then the callable
#must take the force_text parameter.
CODE_TO_MESSAGE_BUILDER_MAP = {
    codes.DUPLICATE_CONSTRAINTS_EXIST: lambda info, force_text:
        "duplicate constraint already exists{0}\n".format(force_text)
        + "\n".join([
            "  " + constraint(info["constraint_type"], constraint_info)
            for constraint_info in info["constraint_info_list"]
        ])
    ,

    codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE: lambda info:
        (
            "{resource_id} is a {parent_type} resource, you should use the"
            " {parent_type} id: {parent_id} when adding constraints"
        ).format(
            **info
        )
    ,
}



from unittest import TestCase

from pcs.lib.booth.config_facade import ConfigFacade
from pcs.lib.booth.config_parser import ConfigItem

### peers

class GetSitesArbitrators(TestCase):
    def test_no_addrs(self):
        sites = []
        arbitrators = []
        conf = ConfigFacade.create(sites, arbitrators)
        self.assertEqual(sites, conf.get_sites())
        self.assertEqual(arbitrators, conf.get_arbitrators())

    def test_sites(self):
        sites = ["1.1.1.1", "2.2.2.2", "3.3.3.3"]
        arbitrators = []
        conf = ConfigFacade.create(sites, arbitrators)
        self.assertEqual(sites, conf.get_sites())
        self.assertEqual(arbitrators, conf.get_arbitrators())

    def test_arbitrators(self):
        sites = []
        arbitrators = ["4.4.4.4", "5.5.5.5"]
        conf = ConfigFacade.create(sites, arbitrators)
        self.assertEqual(sites, conf.get_sites())
        self.assertEqual(arbitrators, conf.get_arbitrators())

    def test_both(self):
        sites = ["1.1.1.1", "2.2.2.2", "3.3.3.3"]
        arbitrators = ["4.4.4.4", "5.5.5.5"]
        conf = ConfigFacade.create(sites, arbitrators)
        self.assertEqual(sites, conf.get_sites())
        self.assertEqual(arbitrators, conf.get_arbitrators())

### tickets

class AddTicket(TestCase):
    def setUp(self):
        self.conf_struct = [
            ConfigItem("site", "site1", []),
            ConfigItem("site", "site2", []),
            ConfigItem("arbitrator", "arbitrator1", []),
            ConfigItem("ticket", "ticketA", []),
        ]
        self.conf = ConfigFacade(self.conf_struct[:])

    def test_add_no_options(self):
        self.conf.add_ticket("ticketB", {})
        self.assertEqual(
            self.conf.config,
            self.conf_struct + [
                ConfigItem("ticket", "ticketB", []),
            ]
        )

    def test_add_with_options(self):
        self.conf.add_ticket("ticketB", {"timeout": "10"})
        self.assertEqual(
            self.conf.config,
            self.conf_struct + [
                ConfigItem(
                    "ticket",
                    "ticketB",
                    [ConfigItem("timeout", "10", [])]
                ),
            ]
        )

class HasTicket(TestCase):
    def setUp(self):
        self.conf_struct = [
            ConfigItem("site", "site1", []),
            ConfigItem("site", "site2", []),
            ConfigItem("arbitrator", "arbitrator1", []),
            ConfigItem("ticket", "ticketA", []),
            ConfigItem(
                "ticket",
                "ticketB",
                [ConfigItem("timeout", "10", [])]
            ),
        ]
        self.conf = ConfigFacade(self.conf_struct[:])

    def test_ticket_exists(self):
        self.assertTrue(self.conf.has_ticket("ticketA"))
        self.assertTrue(self.conf.has_ticket("ticketB"))

    def test_ticket_missing(self):
        self.assertFalse(self.conf.has_ticket("ticketX"))

    def test_not_a_ticket(self):
        self.assertFalse(self.conf.has_ticket("site1"))
        self.assertFalse(self.conf.has_ticket("arbitrator1"))
        self.assertFalse(self.conf.has_ticket("timeout"))

class RemoveTicket(TestCase):
    def setUp(self):
        self.conf_struct_base = [
            ConfigItem("site", "site1", []),
            ConfigItem("site", "site2", []),
            ConfigItem("arbitrator", "arbitrator1", []),
        ]
        self.conf_struct_ticket_a = [
            ConfigItem("ticket", "ticketA", []),
        ]
        self.conf_struct_ticket_b = [
            ConfigItem("ticket", "ticketB", []),
        ]
        self.conf_struct = (
            self.conf_struct_base
            +
            self.conf_struct_ticket_a
            +
            self.conf_struct_ticket_b
        )
        self.conf = ConfigFacade(self.conf_struct[:])

    def test_existing_ticket(self):
        self.conf.remove_ticket("ticketA")
        self.assertEqual(
            self.conf.config,
            self.conf_struct_base + self.conf_struct_ticket_b
        )

    def test_missing_ticket(self):
        self.conf.remove_ticket("ticketX")
        self.assertEqual(self.conf.config, self.conf_struct)

    def test_not_a_ticket(self):
        self.conf.remove_ticket("site1")
        self.assertEqual(self.conf.config, self.conf_struct)

### authfile

class SetAuthfile(TestCase):
    def setUp(self):
        self.conf_struct_base = [
            ConfigItem("site", "site1", []),
            ConfigItem("site", "site2", []),
            ConfigItem("arbitrator", "arbitrator1", []),
        ]
        self.conf_struct_tickets = [
            ConfigItem("ticket", "ticketA", []),
            ConfigItem("ticket", "ticketB", []),
        ]

    def test_add_authfile(self):
        conf = ConfigFacade(
            self.conf_struct_base
            +
            self.conf_struct_tickets
        )
        conf.set_authfile("/path/to/auth.file")
        self.assertEqual(
            conf.config,
            (
                [ConfigItem("authfile", "/path/to/auth.file", [])]
                +
                self.conf_struct_base
                +
                self.conf_struct_tickets
            )
        )

    def test_change_authfile(self):
        conf = ConfigFacade(
            self.conf_struct_base
            +
            [
                ConfigItem("authfile", "/old/path/to/auth1.file", []),
                ConfigItem("authfile", "/old/path/to/auth2.file", []),
            ]
            +
            self.conf_struct_tickets
        )
        conf.set_authfile("/path/to/auth.file")
        self.assertEqual(
            conf.config,
            (
                [ConfigItem("authfile", "/path/to/auth.file", [])]
                +
                self.conf_struct_base
                +
                self.conf_struct_tickets
            )
        )



import sys
import inspect
from functools import partial

from pcs.cli.booth.console_report import (
    CODE_TO_MESSAGE_BUILDER_MAP as BOOTH_CODE_TO_MESSAGE_BUILDER_MAP
)
from pcs.cli.common.console_report import (
    CODE_TO_MESSAGE_BUILDER_MAP,
    error,
    warn,
)
from pcs.cli.constraint_all.console_report import (
    CODE_TO_MESSAGE_BUILDER_MAP as CONSTRAINT_CODE_TO_MESSAGE_BUILDER_MAP
)
from pcs.common import report_codes as codes
from pcs.common.reports import SimpleReportProcessorInterface
from pcs.lib.errors import LibraryError, ReportItemSeverity


CODE_BUILDER_MAP = {}
CODE_BUILDER_MAP.update(CODE_TO_MESSAGE_BUILDER_MAP)
CODE_BUILDER_MAP.update(CONSTRAINT_CODE_TO_MESSAGE_BUILDER_MAP)
CODE_BUILDER_MAP.update(BOOTH_CODE_TO_MESSAGE_BUILDER_MAP)

def build_default_message_from_report(report_item, force_text):
    return "Unknown report: {0} info: {1}{2}".format(
        report_item.code,
        str(report_item.info),
        force_text,
    )


def build_message_from_report(code_builder_map, report_item, force_text=""):
    if report_item.code not in code_builder_map:
        return build_default_message_from_report(report_item, force_text)

    message = code_builder_map[report_item.code]
    #Sometimes report item info is not needed for message building.
    #In that case the message is a string. Otherwise the message is a callable.
    if not callable(message):
        return message + force_text

    try:
        # Object functools.partial cannot be used with inspect because it is not
        # regular python function. We have to use original function for that.
        if isinstance(message, partial):
            keywords = message.keywords if message.keywords is not None else {}
            args = inspect.getfullargspec(message.func).args
            del args[:len(message.args)]
            args = [arg for arg in args if arg not in keywords]
        else:
            args = inspect.getfullargspec(message).args
        if "force_text" in args:
            return message(report_item.info, force_text)
        return message(report_item.info) + force_text
    except(TypeError, KeyError):
        return build_default_message_from_report(report_item, force_text)



build_report_message = partial(build_message_from_report, CODE_BUILDER_MAP)


class LibraryReportProcessorToConsole(SimpleReportProcessorInterface):
    def __init__(self, debug=False):
        self.debug = debug
        self.items = []

    def append(self, report_item):
        self.items.append(report_item)
        return self

    def extend(self, report_item_list):
        self.items.extend(report_item_list)
        return self

    @property
    def errors_count(self):
        return len([
            item for item in self.items
            if item.severity == ReportItemSeverity.ERROR
        ])

    def report_list(self, report_list):
        return self._send(report_list)

    def process(self, report_item):
        self.append(report_item)
        self.send()

    def process_list(self, report_item_list):
        self.extend(report_item_list)
        self.send()

    def _send(self, report_item_list, print_errors=True):
        errors = []
        for report_item in report_item_list:
            if report_item.severity == ReportItemSeverity.ERROR:
                if print_errors:
                    error(build_report_message(
                        report_item,
                        _prepare_force_text(report_item)
                    ))
                errors.append(report_item)
            elif report_item.severity == ReportItemSeverity.WARNING:
                warn(build_report_message(report_item))
            elif self.debug or report_item.severity != ReportItemSeverity.DEBUG:
                msg = build_report_message(report_item)
                if msg:
                    print(msg)
        return errors

    def send(self):
        errors = self._send(self.items, print_errors=False)
        self.items = []
        if errors:
            raise LibraryError(*errors)


def _prepare_force_text(report_item):
    force_text_map = {
        codes.SKIP_OFFLINE_NODES: ", use --skip-offline to override",
    }
    if report_item.forceable:
        return force_text_map.get(
            report_item.forceable, ", use --force to override"
        )
    return ""

def process_library_reports(report_item_list):
    """
    report_item_list list of ReportItem
    """
    if not report_item_list:
        raise error("Errors have occurred, therefore pcs is unable to continue")

    critical_error = False
    for report_item in report_item_list:
        if report_item.severity == ReportItemSeverity.WARNING:
            print("Warning: " + build_report_message(report_item))
            continue

        if report_item.severity != ReportItemSeverity.ERROR:
            print(build_report_message(report_item))
            continue

        sys.stderr.write('Error: {0}\n'.format(build_report_message(
            report_item,
            _prepare_force_text(report_item)
        )))
        critical_error = True

    if critical_error:
        sys.exit(1)



import inspect

from pcs_test.tools.command_env.calls import CallListBuilder
from pcs_test.tools.command_env.config_corosync_conf import CorosyncConf
from pcs_test.tools.command_env.config_env import EnvConfig
from pcs_test.tools.command_env.config_fs import FsConfig
from pcs_test.tools.command_env.config_http import HttpConfig
from pcs_test.tools.command_env.config_raw_file import RawFileConfig
from pcs_test.tools.command_env.config_runner import RunnerConfig

class Spy:
    def __init__(self, known_hosts):
        self.known_hosts = known_hosts

class Config:
    # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.__calls = CallListBuilder()
        self.runner = self.__wrap_helper(
            RunnerConfig(self.__calls, self.__wrap_helper)
        )
        self.env = self.__wrap_helper(EnvConfig(self.__calls))
        self.http = self.__wrap_helper(
            HttpConfig(self.__calls, self.__wrap_helper)
        )
        self.corosync_conf = self.__wrap_helper(CorosyncConf(self.__calls))
        self.fs = self.__wrap_helper(FsConfig(self.__calls))
        self.raw_file = self.__wrap_helper(RawFileConfig(self.__calls))

        self.spy = None

    def add_extension(self, name, Extension):
        if hasattr(self, name):
            raise AssertionError(
                "Config (integration tests) has the extension '{0}' already."
                .format(name)
            )
        setattr(self, name, self.__wrap_helper(
            Extension(self.__calls, self.__wrap_helper, self)
        ))

    def set_spy(self, known_hosts):
        self.spy = Spy(known_hosts)
        return self

    @property
    def calls(self):
        return self.__calls

    def remove(self, name):
        """
        Remove call with specified name from list.
        """
        self.__calls.remove(name)
        return self

    def __wrap_method(self, helper, name, method):
        """
        Wrap method in helper to return self of this object

        object helper -- helper for creatig call configuration
        string name -- name of method in helper
        callable method
        """
        def wrapped_method(*args, **kwargs):
            method(helper, *args, **kwargs)
            return self
        setattr(helper, name, wrapped_method)

    def __wrap_helper(self, helper):
        """
        Wrap every public method in helper to return self of this object

        object helper -- helper for creatig call configuration
        """
        for name, attr in inspect.getmembers(helper.__class__):
            if not name.startswith("_") and hasattr(attr, "__call__"):
                self.__wrap_method(helper, name, attr)
        return helper



from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import(
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common import report_codes
from pcs.lib.commands.constraint import common as constraint
from pcs.lib.errors import ReportItemSeverity as severities


def fixture_cib_and_constraints():
    cib = etree.Element("cib")
    resources_section = etree.SubElement(cib, "resources")
    for _id in ("A", "B", "E", "F"):
        etree.SubElement(resources_section, "primitive").attrib["id"] = _id
    constraint_section = etree.SubElement(
        etree.SubElement(cib, "configuration"),
        "constraints"
    )
    return cib, constraint_section

def fixture_env(cib):
    env = mock.MagicMock()
    env.get_cib = mock.Mock()
    env.get_cib.return_value = cib
    env.push_cib = mock.Mock()
    env.report_processor = MockLibraryReportProcessor()
    return env

class CreateWithSetTest(TestCase):
    def setUp(self):
        self.cib, self.constraint_section = fixture_cib_and_constraints()
        self.env = fixture_env(self.cib)
        self.independent_cib = etree.XML(etree.tostring(self.cib))

    def create(self, duplication_alowed=False):
        constraint.create_with_set(
            "rsc_some",
            lambda cib, options, resource_set_list: options,
            self.env,
            [
                {"ids": ["A", "B"], "options": {"role": "Master"}},
                {"ids": ["E", "F"], "options": {"action": "start"}},
            ],
            {"id":"some_id", "symmetrical": "true"},
            duplication_alowed=duplication_alowed
        )

    def test_put_new_constraint_to_constraint_section(self):
        self.create()
        self.env.push_cib.assert_called_once_with()
        self.independent_cib.find(".//constraints").append(etree.XML("""
            <rsc_some id="some_id" symmetrical="true">
                  <resource_set id="pcs_rsc_set_A_B" role="Master">
                      <resource_ref id="A"></resource_ref>
                      <resource_ref id="B"></resource_ref>
                  </resource_set>
                  <resource_set action="start" id="pcs_rsc_set_E_F">
                      <resource_ref id="E"></resource_ref>
                      <resource_ref id="F"></resource_ref>
                  </resource_set>
            </rsc_some>
        """))
        assert_xml_equal(
            etree.tostring(self.independent_cib).decode(),
            etree.tostring(self.cib).decode()
        )

    def test_refuse_duplicate(self):
        self.create()
        self.env.push_cib.assert_called_once_with()
        assert_raise_library_error(self.create, (
            severities.ERROR,
            report_codes.DUPLICATE_CONSTRAINTS_EXIST,
            {
                'constraint_type': 'rsc_some',
                'constraint_info_list': [{
                    'options': {'symmetrical': 'true', 'id': 'some_id'},
                    'resource_sets': [
                        {
                            'ids': ['A', 'B'],
                            'options':{'role':'Master', 'id':'pcs_rsc_set_A_B'}
                        },
                        {
                            'ids': ['E', 'F'],
                            'options':{'action':'start', 'id':'pcs_rsc_set_E_F'}
                        }
                    ],
                }]
            },
            report_codes.FORCE_CONSTRAINT_DUPLICATE
        ))

    def test_put_duplicate_constraint_when_duplication_allowed(self):
        self.create()
        self.create(duplication_alowed=True)
        expected_calls = [
            mock.call(),
            mock.call(),
        ]
        self.assertEqual(self.env.push_cib.call_count, len(expected_calls))
        self.env.push_cib.assert_has_calls(expected_calls)

        constraint_section = self.independent_cib.find(".//constraints")
        constraint_section.append(etree.XML("""
            <rsc_some id="some_id" symmetrical="true">
                <resource_set id="pcs_rsc_set_A_B" role="Master">
                    <resource_ref id="A"></resource_ref>
                    <resource_ref id="B"></resource_ref>
                </resource_set>
                <resource_set action="start" id="pcs_rsc_set_E_F">
                    <resource_ref id="E"></resource_ref>
                    <resource_ref id="F"></resource_ref>
                </resource_set>
            </rsc_some>
        """))
        constraint_section.append(etree.XML("""
            <rsc_some id="some_id" symmetrical="true">
                <resource_set id="pcs_rsc_set_A_B-1" role="Master">
                    <resource_ref id="A"></resource_ref>
                    <resource_ref id="B"></resource_ref>
                </resource_set>
                <resource_set action="start" id="pcs_rsc_set_E_F-1">
                    <resource_ref id="E"></resource_ref>
                    <resource_ref id="F"></resource_ref>
                </resource_set>
            </rsc_some>
        """))
        assert_xml_equal(
            etree.tostring(self.independent_cib).decode(),
            etree.tostring(self.cib).decode()
        )

class ShowTest(TestCase):
    def setUp(self):
        self.cib, self.constraint_section = fixture_cib_and_constraints()
        self.env = fixture_env(self.cib)

    def create(self, tag_name, resource_set_list):
        constraint.create_with_set(
            tag_name,
            lambda cib, options, resource_set_list: options,
            self.env,
            resource_set_list,
            {"id":"some_id", "symmetrical": "true"},
        )

    def test_returns_export_of_found_elements(self):
        tag_name = "rsc_some"
        self.create(tag_name, [
            {"ids": ["A", "B"], "options": {"role": "Master"}},
        ])
        self.create(tag_name, [
            {"ids": ["E", "F"], "options": {"action": "start"}},
        ])
        etree.SubElement(self.constraint_section, tag_name).attrib.update({
            "id": "plain1", "is_plain": "true"
        })

        is_plain = lambda element: element.attrib.has_key("is_plain")

        self.assertEqual(
            constraint.show(tag_name, is_plain, self.env), {
            'plain': [{"options": {'id': 'plain1', 'is_plain': 'true'}}],
            'with_resource_sets': [
                {
                    'resource_sets': [{
                        'ids': ['A', 'B'],
                        'options': {'role': 'Master', 'id': 'pcs_rsc_set_A_B'},
                    }],
                    'options': {'symmetrical': 'true', 'id': 'some_id'}
                },
                {
                    'options': {'symmetrical': 'true', 'id': 'some_id'},
                    'resource_sets': [{
                        'ids': ['E', 'F'],
                        'options': {'action': 'start', 'id': 'pcs_rsc_set_E_F'}
                    }]
                }
            ]
        })



import json

from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call
)

class PcmkShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def set_stonith_watchdog_timeout_to_zero(
        self, node_labels=None, communication_list=None,
        name="http.pcmk.set_stonith_watchdog_timeout_to_zero"
    ):
        """
        Create a call for setting on hosts

        node_labels list -- create success responses from these nodes
        communication_list list -- create custom responses
        name string -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/set_stonith_watchdog_timeout_to_zero"
        )

    def remove_stonith_watchdog_timeout(
        self, node_labels=None, communication_list=None,
        name="http.pcmk.remove_stonith_watchdog_timeout"
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/remove_stonith_watchdog_timeout"
        )

    def remove_nodes_from_cib(
        self, nodes_to_remove, node_labels=None, communication_list=None,
        name="http.pcmk.remove_nodes_from_cib",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/remove_nodes_from_cib",
            param_list=[
                ("data_json", json.dumps(dict(node_list=nodes_to_remove)))
            ],
            output=json.dumps(dict(
                code="success",
                message="",
            )),
        )



import base64
import os.path
from functools import partial

from pcs import settings
from pcs.common import (
    file_type_codes,
    report_codes,
)
from pcs.common.file import FileAlreadyExists, RawFileError
from pcs.common.reports import SimpleReportProcessor
from pcs.common.tools import join_multilines
from pcs.lib import external, reports, tools
from pcs.lib.cib.resource import primitive, group
from pcs.lib.booth import (
    config_files,
    config_validators,
    constants,
    reports as booth_reports,
    resource,
    status,
)
from pcs.lib.cib.tools import get_resources, IdProvider
from pcs.lib.communication.booth import (
    BoothGetConfig,
    BoothSendConfig,
)
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.errors import LibraryError, ReportItemSeverity
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import GhostFile, raw_file_error_report
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.resource_agent import find_valid_resource_agent_by_name


def config_setup(
    env, site_list, arbitrator_list, instance_name=None,
    overwrite_existing=False
):
    """
    create booth configuration

    LibraryEnvironment env
    list site_list -- site adresses of multisite
    list arbitrator_list -- arbitrator adresses of multisite
    string instance_name -- booth instance name
    bool overwrite_existing -- allow overwriting existing files
    """
    instance_name = instance_name or constants.DEFAULT_INSTANCE_NAME
    report_processor = SimpleReportProcessor(env.report_processor)

    report_processor.report_list(
        config_validators.check_instance_name(instance_name)
    )
    report_processor.report_list(
        config_validators.create(site_list, arbitrator_list)
    )
    if report_processor.has_errors:
        raise LibraryError()

    booth_env = env.get_booth_env(instance_name)

    booth_conf = booth_env.create_facade(site_list, arbitrator_list)
    booth_conf.set_authfile(booth_env.key_path)

    report_creator = reports.get_problem_creator(
        force_code=report_codes.FORCE_FILE_OVERWRITE,
        is_forced=overwrite_existing
    )
    try:
        booth_env.key.write_raw(
            tools.generate_binary_key(
                random_bytes_count=settings.booth_authkey_bytes
            ),
            can_overwrite=overwrite_existing
        )
        booth_env.config.write_facade(
            booth_conf,
            can_overwrite=overwrite_existing
        )
    except FileAlreadyExists as e:
        report_processor.report(
            report_creator(
                reports.file_already_exists,
                e.metadata.file_type_code,
                e.metadata.path,
            )
        )
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    if report_processor.has_errors:
        raise LibraryError()


def config_destroy(env, instance_name=None, ignore_config_load_problems=False):
    # pylint: disable=too-many-branches
    """
    remove booth configuration files

    LibraryEnvironment env
    string instance_name -- booth instance name
    bool ignore_config_load_problems -- delete as much as possible when unable
            to read booth configs for the given booth instance
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    instance_name = booth_env.instance_name
    _ensure_live_env(env, booth_env)

    # TODO use constants in reports
    config_is_used = partial(booth_reports.booth_config_is_used, instance_name)
    if resource.find_for_config(
        get_resources(env.get_cib()),
        booth_env.config_path,
    ):
        report_processor.report(config_is_used("in cluster resource"))
    # Only systemd is currently supported. Initd does not supports multiple
    # instances (here specified by name)
    if external.is_systemctl():
        if external.is_service_running(
            env.cmd_runner(), "booth", instance_name
        ):
            report_processor.report(config_is_used("(running in systemd)"))

        if external.is_service_enabled(
            env.cmd_runner(), "booth", instance_name
        ):
            report_processor.report(config_is_used("(enabled in systemd)"))
    if report_processor.has_errors:
        raise LibraryError()

    try:
        authfile_path = None
        booth_conf = booth_env.config.read_to_facade()
        authfile_path = booth_conf.get_authfile()
    except RawFileError as e:
        report_processor.report(
            raw_file_error_report(
                e,
                force_code=report_codes.FORCE_BOOTH_DESTROY,
                is_forced_or_warning=ignore_config_load_problems,
            )
        )
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(
                e,
                force_code=report_codes.FORCE_BOOTH_DESTROY,
                is_forced_or_warning=ignore_config_load_problems,
            )
        )
    if report_processor.has_errors:
        raise LibraryError()

    if authfile_path:
        authfile_dir, authfile_name = os.path.split(authfile_path)
        if (authfile_dir == settings.booth_config_dir) and authfile_name:
            try:
                key_file = FileInstance.for_booth_key(authfile_name)
                key_file.raw_file.remove(fail_if_file_not_found=False)
            except RawFileError as e:
                report_processor.report(
                    raw_file_error_report(
                        e,
                        force_code=report_codes.FORCE_BOOTH_DESTROY,
                        is_forced_or_warning=ignore_config_load_problems,
                    )
                )
        else:
            report_processor.report(
                booth_reports.booth_unsupported_file_location(
                    authfile_path,
                    settings.booth_config_dir,
                    file_type_codes.BOOTH_KEY,
                )
            )
    if report_processor.has_errors:
        raise LibraryError()

    try:
        booth_env.config.raw_file.remove()
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))

    if report_processor.has_errors:
        raise LibraryError()


def config_text(env, instance_name=None, node_name=None):
    """
    get configuration in raw format

    string instance_name -- booth instance name
    string node_name -- get the config from specified node or local host if None
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    instance_name = booth_env.instance_name
    # It does not make any sense for the cli to read a ghost file and send it
    # to lib so that the lib could return it unchanged to cli. Just use 'cat'.
    # When node_name is specified, using ghost files doesn't make any sense
    # either.
    _ensure_live_env(env, booth_env)

    if node_name is None:
        try:
            return booth_env.config.read_raw()
        except RawFileError as e:
            report_processor.report(raw_file_error_report(e))
        if report_processor.has_errors:
            raise LibraryError()

    com_cmd = BoothGetConfig(env.report_processor, instance_name)
    com_cmd.set_targets([
        env.get_node_target_factory().get_target_from_hostname(node_name)
    ])
    # pylint: disable=unsubscriptable-object
    # In general, pylint is right. And it cannot know in this case code is OK.
    # It is covered by tests.
    remote_data = run_and_raise(env.get_node_communicator(), com_cmd)[0][1]
    try:
        # TODO switch to new file transfer commands (not implemented yet)
        # which send and receive configs as bytes instead of strings
        return remote_data["config"]["data"].encode("utf-8")
    except KeyError:
        raise LibraryError(reports.invalid_response_format(node_name))


def config_ticket_add(
    env, ticket_name, options, instance_name=None, allow_unknown_options=False
):
    """
    add a ticket to booth configuration

    LibraryEnvironment env
    string ticket_name -- the name of the ticket to be created
    dict options -- options for the ticket
    string instance_name -- booth instance name
    bool allow_unknown_options -- allow using options unknown to pcs
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    try:
        booth_conf = booth_env.config.read_to_facade()
        report_processor.report_list(
            config_validators.add_ticket(
                booth_conf,
                ticket_name,
                options,
                allow_unknown_options=allow_unknown_options
            )
        )
        if report_processor.has_errors:
            raise LibraryError()
        booth_conf.add_ticket(ticket_name, options)
        booth_env.config.write_facade(booth_conf, can_overwrite=True)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()


def config_ticket_remove(env, ticket_name, instance_name=None):
    """
    remove a ticket from booth configuration

    LibraryEnvironment env
    string ticket_name -- the name of the ticket to be removed
    string instance_name -- booth instance name
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    try:
        booth_conf = booth_env.config.read_to_facade()
        report_processor.report_list(
            config_validators.remove_ticket(booth_conf, ticket_name)
        )
        if report_processor.has_errors:
            raise LibraryError()
        booth_conf.remove_ticket(ticket_name)
        booth_env.config.write_facade(booth_conf, can_overwrite=True)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()


def create_in_cluster(
    env, ip, instance_name=None, allow_absent_resource_agent=False
):
    """
    Create group with ip resource and booth resource

    LibraryEnvironment env -- provides all for communication with externals
    string ip -- float ip address for the operation of the booth
    string instance_name -- booth instance name
    bool allow_absent_resource_agent -- allowing creating booth resource even
        if its agent is not installed
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    # Booth config path goes to CIB. Working with a mocked booth configs would
    # not work coorectly as the path would point to a mock file (the path to a
    # mock file is unknown to us in the lib anyway)
    # It makes sense to work with a mocked CIB, though. Users can do other
    # changes to the CIB and push them to the cluster at once.
    _ensure_live_booth_env(booth_env)
    resources_section = get_resources(env.get_cib())
    id_provider = IdProvider(resources_section)
    instance_name = booth_env.instance_name

    # validate
    if resource.find_for_config(resources_section, booth_env.config_path):
        report_processor.report(
            booth_reports.booth_already_in_cib(instance_name)
        )
    # verify the config exists and is readable
    try:
        booth_env.config.raw_file.read()
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    if report_processor.has_errors:
        raise LibraryError()
    # validation done

    create_id = partial(
        resource.create_resource_id,
        resources_section,
        instance_name
    )
    get_agent = partial(
        find_valid_resource_agent_by_name,
        env.report_processor,
        env.cmd_runner(),
        allowed_absent=allow_absent_resource_agent
    )
    create_primitive = partial(
        primitive.create,
        env.report_processor,
        resources_section,
        id_provider
    )
    into_booth_group = partial(
        group.place_resource,
        group.provide_group(resources_section, create_id("group")),
    )

    into_booth_group(create_primitive(
        create_id("ip"),
        get_agent("ocf:heartbeat:IPaddr2"),
        instance_attributes={"ip": ip},
    ))
    into_booth_group(create_primitive(
        create_id("service"),
        get_agent("ocf:pacemaker:booth-site"),
        instance_attributes={"config": booth_env.config_path},
    ))

    env.push_cib()


def remove_from_cluster(
    env, resource_remove, instance_name=None, allow_remove_multiple=False
):
    """
    Remove group with ip resource and booth resource

    LibraryEnvironment env -- provides all for communication with externals
    function resource_remove -- provisional hack til resources are moved to lib
    string instance_name -- booth instance name
    bool allow_remove_multiple -- remove all resources if more than one found
    """
    # TODO resource_remove is provisional hack til resources are moved to lib
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    # This command does not work with booth config files at all, let's reject
    # them then.
    _ensure_live_booth_env(booth_env)

    resource.get_remover(resource_remove)(
        _find_resource_elements_for_operation(
            report_processor,
            get_resources(env.get_cib()),
            booth_env,
            allow_remove_multiple,
        )
    )


def restart(env, resource_restart, instance_name=None, allow_multiple=False):
    """
    Restart group with ip resource and booth resource

    LibraryEnvironment env -- provides all for communication with externals
    function resource_restart -- provisional hack til resources are moved to lib
    string instance_name -- booth instance name
    bool allow_remove_multiple -- remove all resources if more than one found
    """
    # TODO resource_remove is provisional hack til resources are moved to lib
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)

    for booth_element in _find_resource_elements_for_operation(
        report_processor,
        get_resources(env.get_cib()),
        booth_env,
        allow_multiple,
    ):
        resource_restart([booth_element.attrib["id"]])


def ticket_grant(env, ticket_name, site_ip=None, instance_name=None):
    """
    Grant a ticket to the site specified by site_ip

    LibraryEnvironment env
    string ticket_name -- the name of the ticket to be granted
    string site_ip -- IP of the site to grant the ticket to, None for local
    string instance_name -- booth instance name
    """
    return _ticket_operation(
        "grant",
        env,
        ticket_name,
        site_ip=site_ip,
        instance_name=instance_name,
    )


def ticket_revoke(env, ticket_name, site_ip=None, instance_name=None):
    """
    Revoke a ticket from the site specified by site_ip

    LibraryEnvironment env
    string ticket_name -- the name of the ticket to be revoked
    string site_ip -- IP of the site to revoke the ticket from, None for local
    string instance_name -- booth instance name
    """
    return _ticket_operation(
        "revoke",
        env,
        ticket_name,
        site_ip=site_ip,
        instance_name=instance_name,
    )


def _ticket_operation(operation, env, ticket_name, site_ip, instance_name):
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)

    if not site_ip:
        site_ip_list = resource.find_bound_ip(
            get_resources(env.get_cib()),
            booth_env.config_path
        )
        if len(site_ip_list) != 1:
            raise LibraryError(
                booth_reports.booth_cannot_determine_local_site_ip()
            )
        site_ip = site_ip_list[0]

    stdout, stderr, return_code = env.cmd_runner().run([
        settings.booth_binary, operation, "-s", site_ip, ticket_name
    ])

    if return_code != 0:
        raise LibraryError(
            booth_reports.booth_ticket_operation_failed(
                operation,
                join_multilines([stderr, stdout]),
                site_ip,
                ticket_name
            )
        )


def config_sync(env, instance_name=None, skip_offline_nodes=False):
    """
    Send specified local booth configuration to all nodes in the local cluster.

    LibraryEnvironment env
    string instance_name -- booth instance name
    skip_offline_nodes -- if True offline nodes will be skipped
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    if not env.is_cib_live:
        raise LibraryError(
            reports.live_environment_required(
                [file_type_codes.CIB],
            )
        )

    cluster_nodes_names, report_list = get_existing_nodes_names(
        env.get_corosync_conf()
    )
    if not cluster_nodes_names:
        report_list.append(reports.corosync_config_no_nodes_defined())
    report_processor.report_list(report_list)

    try:
        booth_conf_data = booth_env.config.read_raw()
        booth_conf = booth_env.config.raw_to_facade(booth_conf_data)
        if isinstance(booth_env.config.raw_file, GhostFile):
            authfile_data = booth_env.key.read_raw()
            authfile_path = booth_conf.get_authfile()
            authfile_name = (
                os.path.basename(authfile_path) if authfile_path else None
            )
        else:
            authfile_name, authfile_data, authfile_report_list = (
                config_files.get_authfile_name_and_data(booth_conf)
            )
            report_processor.report_list(authfile_report_list)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            booth_env.config.parser_exception_to_report_list(e)
        )
    if report_processor.has_errors:
        raise LibraryError()

    com_cmd = BoothSendConfig(
        env.report_processor,
        booth_env.instance_name,
        booth_conf_data,
        authfile=authfile_name,
        authfile_data=authfile_data,
        skip_offline_targets=skip_offline_nodes
    )
    com_cmd.set_targets(
        env.get_node_target_factory().get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
    )
    run_and_raise(env.get_node_communicator(), com_cmd)


def enable_booth(env, instance_name=None):
    """
    Enable specified instance of booth service, systemd systems supported only.

    LibraryEnvironment env
    string instance_name -- booth instance name
    """
    external.ensure_is_systemd()
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        external.enable_service(env.cmd_runner(), "booth", instance_name)
    except external.EnableServiceError as e:
        raise LibraryError(reports.service_enable_error(
            "booth", e.message, instance=instance_name
        ))
    env.report_processor.process(reports.service_enable_success(
        "booth", instance=instance_name
    ))


def disable_booth(env, instance_name=None):
    """
    Disable specified instance of booth service, systemd systems supported only.

    LibraryEnvironment env
    string instance_name -- booth instance name
    """
    external.ensure_is_systemd()
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        external.disable_service(env.cmd_runner(), "booth", instance_name)
    except external.DisableServiceError as e:
        raise LibraryError(reports.service_disable_error(
            "booth", e.message, instance=instance_name
        ))
    env.report_processor.process(reports.service_disable_success(
        "booth", instance=instance_name
    ))


def start_booth(env, instance_name=None):
    """
    Start specified instance of booth service, systemd systems supported only.
        On non-systemd systems it can be run like this:
        BOOTH_CONF_FILE=<booth-file-path> /etc/initd/booth-arbitrator

    LibraryEnvironment env
    string instance_name -- booth instance name
    """
    external.ensure_is_systemd()
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        external.start_service(env.cmd_runner(), "booth", instance_name)
    except external.StartServiceError as e:
        raise LibraryError(reports.service_start_error(
            "booth", e.message, instance=instance_name
        ))
    env.report_processor.process(reports.service_start_success(
        "booth", instance=instance_name
    ))


def stop_booth(env, instance_name=None):
    """
    Stop specified instance of booth service, systemd systems supported only.

    LibraryEnvironment env
    string instance_name -- booth instance name
    """
    external.ensure_is_systemd()
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name

    try:
        external.stop_service(env.cmd_runner(), "booth", instance_name)
    except external.StopServiceError as e:
        raise LibraryError(reports.service_stop_error(
            "booth", e.message, instance=instance_name
        ))
    env.report_processor.process(reports.service_stop_success(
        "booth", instance=instance_name
    ))


def pull_config(env, node_name, instance_name=None):
    """
    Get config from specified node and save it on local system. It will
    rewrite existing files.

    LibraryEnvironment env
    string node_name -- name of the node from which the config should be fetched
    string instance_name -- booth instance name
    """
    report_processor = SimpleReportProcessor(env.report_processor)
    booth_env = env.get_booth_env(instance_name)
    instance_name = booth_env.instance_name
    _ensure_live_env(env, booth_env)

    env.report_processor.process(
        booth_reports.booth_fetching_config_from_node_started(
            node_name, instance_name
        )
    )
    com_cmd = BoothGetConfig(env.report_processor, instance_name)
    com_cmd.set_targets([
        env.get_node_target_factory().get_target_from_hostname(node_name)
    ])
    # pylint: disable=unsubscriptable-object
    # In general, pylint is right. And it cannot know in this case code is OK.
    # It is covered by tests.
    output = run_and_raise(env.get_node_communicator(), com_cmd)[0][1]
    try:
        # TODO adapt to new file transfer framework once it is written
        if (
            output["authfile"]["name"] is not None
            and
            output["authfile"]["data"]
        ):
            authfile_name = output["authfile"]["name"]
            report_list = config_validators.check_instance_name(authfile_name)
            if report_list:
                raise LibraryError(*report_list)
            booth_key = FileInstance.for_booth_key(authfile_name)
            booth_key.write_raw(
                base64.b64decode(
                    output["authfile"]["data"].encode("utf-8")
                ),
                can_overwrite=True
            )
        booth_env.config.write_raw(
            output["config"]["data"].encode("utf-8"),
            can_overwrite=True
        )
        env.report_processor.process(
            booth_reports.booth_config_accepted_by_node(
                name_list=[instance_name]
            )
        )
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except KeyError:
        raise LibraryError(reports.invalid_response_format(node_name))
    if report_processor.has_errors:
        raise LibraryError()


def get_status(env, instance_name=None):
    """
    get booth status info

    LibraryEnvironment env
    string instance_name -- booth instance name
    """
    booth_env = env.get_booth_env(instance_name)
    _ensure_live_env(env, booth_env)
    instance_name = booth_env.instance_name
    return {
        "status": status.get_daemon_status(env.cmd_runner(), instance_name),
        "ticket": status.get_tickets_status(env.cmd_runner(), instance_name),
        "peers": status.get_peers_status(env.cmd_runner(), instance_name),
    }


def _find_resource_elements_for_operation(
    report_processor, resources_section, booth_env, allow_multiple
):
    booth_element_list = resource.find_for_config(
        resources_section,
        booth_env.config_path,
    )

    if not booth_element_list:
        report_processor.report(
            booth_reports.booth_not_exists_in_cib(booth_env.instance_name)
        )
    elif len(booth_element_list) > 1:
        report_processor.report(
            booth_reports.booth_multiple_times_in_cib(
                booth_env.instance_name,
                severity=(
                    ReportItemSeverity.WARNING if allow_multiple
                    else ReportItemSeverity.ERROR
                )
            )
        )
    if report_processor.has_errors:
        raise LibraryError()

    return booth_element_list

def _ensure_live_booth_env(booth_env):
    if booth_env.ghost_file_codes:
        raise LibraryError(
            reports.live_environment_required(booth_env.ghost_file_codes)
        )

def _ensure_live_env(env, booth_env):
    not_live = (
        booth_env.ghost_file_codes
        +
        # parenthesis are cruciual, otherwise the if..else influences
        # booth_env.ghost_file_codes as well
        ([file_type_codes.CIB] if not env.is_cib_live else [])
    )
    if not_live:
        raise LibraryError(reports.live_environment_required(not_live))



from pcs import (
    quorum,
    usage,
)
from pcs.cli.common.routing import create_router


quorum_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.quorum(argv),
        "config": quorum.quorum_config_cmd,
        "expected-votes": quorum.quorum_expected_votes_cmd,
        "status": quorum.quorum_status_cmd,
        "device": create_router(
            {
                "add": quorum.quorum_device_add_cmd,
                "heuristics": create_router(
                    {
                        "delete": quorum.quorum_device_heuristics_remove_cmd,
                        "remove": quorum.quorum_device_heuristics_remove_cmd,
                    },
                    ["quorum", "device", "heuristics"]
                ),
                "delete": quorum.quorum_device_remove_cmd,
                "remove": quorum.quorum_device_remove_cmd,
                "status": quorum.quorum_device_status_cmd,
                "update": quorum.quorum_device_update_cmd,
            },
            ["quorum", "device"]
        ),
        # TODO switch to new architecture
        "unblock": quorum.quorum_unblock_cmd,
        "update": quorum.quorum_update_cmd,
    },
    ["quorum"],
    default_cmd="config",
)



import sys
import json
import logging

from typing import (
    List,
    Any,
)

from pcs import settings, utils
from pcs.cli.common.env_cli import Env
from pcs.cli.common.lib_wrapper import Library
from pcs.cli.common.reports import (
    build_report_message,
    LibraryReportProcessorToConsole,
)
from pcs.common import report_codes
from pcs.lib.errors import LibraryError, ReportItemSeverity


SUPPORTED_COMMANDS = {
    "cluster.setup",
    "cluster.add_nodes",
    "cluster.remove_nodes",
}


def _exit(status, status_msg=None, report_list=None, data=None):
    json.dump(
        dict(
            status=status,
            status_msg=status_msg,
            report_list=report_list or [],
            data=data,
        ),
        sys.stdout
    )
    sys.exit(0)


def get_cli_env(options):
    env = Env()
    env.user, env.groups = utils.get_cib_user_groups()
    env.known_hosts_getter = utils.read_known_hosts_file
    # Debug messages always go to the processor. The parameter only affects if
    # they will be printed to stdout. We are not printing the messages. Instead
    # we get all the messages the processor got. So the value of the parameter
    # does not matter.
    env.report_processor = LibraryReportProcessor(True)
    env.request_timeout = (
        options.get("request_timeout") or settings.default_request_timeout
    )
    return env


class LibraryReportProcessor(LibraryReportProcessorToConsole):
    # Note: not properly typed
    processed_items: List[Any] = []
    def _send(self, report_item_list, print_errors=True):

        self.processed_items.extend(report_item_list)
        return [
            report
            for report in report_item_list
            if report.severity == ReportItemSeverity.ERROR
        ]

def export_reports(report_list):
    return [report_item_to_dict(report) for report in report_list]

def report_item_to_dict(report_item):
    return dict(
        severity=report_item.severity,
        code=report_item.code,
        info=report_item.info,
        forceable=None if report_item.forceable is None else report_codes.FORCE,
        report_text=build_report_message(report_item),
    )

def main():
    # pylint: disable=broad-except
    argv = sys.argv[1:]
    if argv:
        _exit("input_error", status_msg="No arguments allowed")

    utils.subprocess_setup()
    logging.basicConfig()

    try:
        input_data = json.load(sys.stdin)
        cli_env = get_cli_env(input_data.get("options", {}))
        lib = Library(cli_env, utils.get_middleware_factory())
        cmd = input_data["cmd"]
        if cmd not in SUPPORTED_COMMANDS:
            _exit("unknown_cmd", status_msg=f"Unknown command '{cmd}'")
        for sub_cmd in cmd.split("."):
            lib = getattr(lib, sub_cmd)
        output_data = lib(**input_data["cmd_data"])
        _exit(
            "success",
            report_list=export_reports(
                cli_env.report_processor.processed_items
            ),
            data=output_data,
        )
    except LibraryError as e:
        _exit(
            "error",
            report_list=export_reports(
                cli_env.report_processor.processed_items + list(e.args)
            ),
        )
    except json.JSONDecodeError as e:
        _exit("input_error", status_msg=f"Unable to parse input data: {e.msg}")
    except KeyError as e:
        _exit("input_error", status_msg=f"Missing key {e}")
    except Exception as e:
        # TODO: maybe add traceback?
        _exit("exception", status_msg=str(e))



from functools import partial
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)

from pcs.common import report_codes
from pcs.lib.cib.constraint import ticket
from pcs.lib.errors import ReportItemSeverity as severities


@mock.patch("pcs.lib.cib.constraint.ticket.tools.check_new_id_applicable")
class PrepareOptionsPlainTest(TestCase):
    def setUp(self):
        self.cib = "cib"
        self.prepare = partial(ticket.prepare_options_plain, self.cib)

    @mock.patch("pcs.lib.cib.constraint.ticket._create_id")
    def test_prepare_correct_options(self, mock_create_id, _):
        mock_create_id.return_value = "generated_id"
        self.assertEqual(
            {
                'id': 'generated_id',
                'loss-policy': 'fence',
                'rsc': 'resourceA',
                'rsc-role': 'Master',
                'ticket': 'ticket_key'
            },
            self.prepare(
                {"loss-policy": "fence", "rsc-role": "master"},
                "ticket_key",
                "resourceA",
            )
        )

    @mock.patch("pcs.lib.cib.constraint.ticket._create_id")
    def test_does_not_include_role_if_not_presented(self, mock_create_id, _):
        mock_create_id.return_value = "generated_id"
        self.assertEqual(
            {
                'id': 'generated_id',
                'loss-policy': 'fence',
                'rsc': 'resourceA',
                'ticket': 'ticket_key'
            },
            self.prepare(
                {"loss-policy": "fence", "rsc-role": ""},
                "ticket_key",
                "resourceA",
            )
        )

    def test_refuse_unknown_attributes(self, _):
        assert_raise_library_error(
            lambda: self.prepare(
                {"unknown": "nonsense", "rsc-role": "master"},
                "ticket_key",
                "resourceA",
            ),
            (
                severities.ERROR,
                report_codes.INVALID_OPTIONS,
                {
                    "option_names": ["unknown"],
                    "option_type": None,
                    "allowed": [
                        "id", "loss-policy", "rsc", "rsc-role", "ticket",
                    ],
                    "allowed_patterns": [],
                }
            ),
        )

    def test_refuse_bad_role(self, _):
        assert_raise_library_error(
            lambda: self.prepare(
                {"id": "id", "rsc-role": "bad_role"}, "ticket_key", "resourceA"
            ),
            (severities.ERROR, report_codes.INVALID_OPTION_VALUE, {
                'allowed_values': ('Master', 'Slave', 'Started', 'Stopped'),
                'option_value': 'bad_role',
                'option_name': 'rsc-role',
                'cannot_be_empty': False,
                'forbidden_characters': None,
            }),
        )

    def test_refuse_missing_ticket(self, _):
        assert_raise_library_error(
            lambda: self.prepare(
                {"id": "id", "rsc-role": "master"}, "", "resourceA"
            ),
            (
                severities.ERROR,
                report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                {
                    "option_names": ["ticket"]
                }
            ),
        )

    def test_refuse_missing_resource_id(self, _):
        assert_raise_library_error(
            lambda: self.prepare(
                {"id": "id", "rsc-role": "master"}, "ticket_key", ""
            ),
            (
                severities.ERROR,
                report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                {
                    "option_names": ["rsc"],
                }
            ),
        )


    def test_refuse_unknown_lost_policy(self, mock_check_new_id_applicable):
        # pylint: disable=unused-argument
        assert_raise_library_error(
            lambda: self.prepare(
                {"loss-policy": "unknown", "ticket": "T", "id": "id"},
                "ticket_key",
                "resourceA",
            ),
            (severities.ERROR, report_codes.INVALID_OPTION_VALUE, {
                'allowed_values': ('fence', 'stop', 'freeze', 'demote'),
                'option_value': 'unknown',
                'option_name': 'loss-policy',
                'cannot_be_empty': False,
                'forbidden_characters': None,
            }),
        )

    @mock.patch("pcs.lib.cib.constraint.ticket._create_id")
    def test_complete_id(self, mock_create_id, _):
        mock_create_id.return_value = "generated_id"
        options = {"loss-policy": "freeze", "ticket": "T", "rsc-role": "Master"}
        ticket_key = "ticket_key"
        resource_id = "resourceA"
        expected_options = options.copy()
        expected_options.update({
            "id": "generated_id",
            "rsc": resource_id,
            "rsc-role": "Master",
            "ticket": ticket_key,
        })
        self.assertEqual(expected_options, self.prepare(
            options,
            ticket_key,
            resource_id,
        ))
        mock_create_id.assert_called_once_with(
            self.cib,
            ticket_key,
            resource_id,
            "Master",
        )


#Patch check_new_id_applicable is always desired when working with
#prepare_options_with_set. Patched function raises when id not applicable
#and do nothing when applicable - in this case tests do no actions with it
@mock.patch("pcs.lib.cib.constraint.ticket.tools.check_new_id_applicable")
class PrepareOptionsWithSetTest(TestCase):
    def setUp(self):
        self.cib = "cib"
        self.resource_set_list = "resource_set_list"
        self.prepare = lambda options: ticket.prepare_options_with_set(
            self.cib,
            options,
            self.resource_set_list,
        )

    @mock.patch("pcs.lib.cib.constraint.ticket.constraint.create_id")
    def test_complete_id(self, mock_create_id, _):
        mock_create_id.return_value = "generated_id"
        options = {"loss-policy": "freeze", "ticket": "T"}
        expected_options = options.copy()
        expected_options.update({"id": "generated_id"})
        self.assertEqual(expected_options, self.prepare(options))
        mock_create_id.assert_called_once_with(
            self.cib,
            ticket.TAG_NAME,
            self.resource_set_list
        )

    def test_refuse_invalid_id(self, mock_check_new_id_applicable):
        class SomeException(Exception):
            pass
        mock_check_new_id_applicable.side_effect = SomeException()
        invalid_id = "invalid_id"
        self.assertRaises(SomeException, lambda: self.prepare({
            "loss-policy": "freeze",
            "ticket": "T",
            "id": invalid_id,
        }))
        mock_check_new_id_applicable.assert_called_once_with(
            self.cib,
            ticket.DESCRIPTION,
            invalid_id
        )

    def test_refuse_unknown_lost_policy(self, _):
        assert_raise_library_error(
            lambda: self.prepare({
                "loss-policy": "unknown",
                "ticket": "T",
                "id": "id",
            }),
            (severities.ERROR, report_codes.INVALID_OPTION_VALUE, {
                'allowed_values': ('fence', 'stop', 'freeze', 'demote'),
                'option_value': 'unknown',
                'option_name': 'loss-policy',
                'cannot_be_empty': False,
                'forbidden_characters': None,
            }),
        )

    def test_refuse_missing_ticket(self, _):
        assert_raise_library_error(
            lambda: self.prepare({"loss-policy": "stop", "id": "id"}),
            (
                severities.ERROR,
                report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                {"option_names": ["ticket"]}
            )
        )

    def test_refuse_empty_ticket(self, _):
        assert_raise_library_error(
            lambda: self.prepare({
                "loss-policy": "stop",
                "id": "id",
                "ticket": " "
            }),
            (
                severities.ERROR,
                report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                {"option_names": ["ticket"]}
            )
        )

class Element:
    # pylint: disable=too-few-public-methods
    def __init__(self, attrib):
        self.attrib = attrib

    def update(self, attrib):
        self.attrib.update(attrib)
        return self


class AreDuplicatePlain(TestCase):
    def setUp(self):
        self.first = Element({
            "ticket": "ticket_key",
            "rsc": "resourceA",
            "rsc-role": "Master"
        })
        self.second = Element({
            "ticket": "ticket_key",
            "rsc": "resourceA",
            "rsc-role": "Master"
        })

    def test_returns_true_for_duplicate_elements(self):
        self.assertTrue(ticket.are_duplicate_plain(self.first, self.second))

    def test_returns_false_for_different_ticket(self):
        self.assertFalse(ticket.are_duplicate_plain(
            self.first,
            self.second.update({"ticket": "X"})
        ))

    def test_returns_false_for_different_resource(self):
        self.assertFalse(ticket.are_duplicate_plain(
            self.first,
            self.second.update({"rsc": "Y"})
        ))

    def test_returns_false_for_different_role(self):
        self.assertFalse(ticket.are_duplicate_plain(
            self.first,
            self.second.update({"rsc-role": "Z"})
        ))

    def test_returns_false_for_different_elements(self):
        self.second.update({
            "ticket": "X",
            "rsc": "Y",
            "rsc-role": "Z"
        })
        self.assertFalse(ticket.are_duplicate_plain(self.first, self.second))

@mock.patch(
    "pcs.lib.cib.constraint.ticket.constraint.have_duplicate_resource_sets"
)
class AreDuplicateWithResourceSet(TestCase):
    def test_returns_true_for_duplicate_elements(
        self, mock_have_duplicate_resource_sets
    ):
        mock_have_duplicate_resource_sets.return_value = True
        self.assertTrue(ticket.are_duplicate_with_resource_set(
            Element({"ticket": "ticket_key"}),
            Element({"ticket": "ticket_key"}),
        ))

    def test_returns_false_for_different_elements(
        self, mock_have_duplicate_resource_sets
    ):
        mock_have_duplicate_resource_sets.return_value = True
        self.assertFalse(ticket.are_duplicate_with_resource_set(
            Element({"ticket": "ticket_key"}),
            Element({"ticket": "X"}),
        ))

class RemovePlainTest(TestCase):
    def test_remove_tickets_constraints_for_resource(self):
        constraint_section = etree.fromstring("""
            <constraints>
                <rsc_ticket id="t1" ticket="tA" rsc="rA"/>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t3" ticket="tA" rsc="rA"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """)

        self.assertTrue(ticket.remove_plain(
            constraint_section,
            ticket_key="tA",
            resource_id="rA",
        ))

        assert_xml_equal(etree.tostring(constraint_section).decode(), """
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """)

    def test_remove_nothing_when_no_matching_found(self):
        constraint_section = etree.fromstring("""
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """)

        self.assertFalse(ticket.remove_plain(
            constraint_section,
            ticket_key="tA",
            resource_id="rA",
        ))

        assert_xml_equal(etree.tostring(constraint_section).decode(), """
            <constraints>
                <rsc_ticket id="t2" ticket="tA" rsc="rB"/>
                <rsc_ticket id="t4" ticket="tB" rsc="rA"/>
                <rsc_ticket id="t5" ticket="tB" rsc="rB"/>
            </constraints>
        """)

class RemoveWithSetTest(TestCase):
    def test_remove_resource_references_and_empty_remaining_parents(self):
        constraint_section = etree.fromstring("""
            <constraints>
                <rsc_ticket id="t1" ticket="tA">
                    <resource_set id="rs1">
                        <resource_ref id="rA"/>
                    </resource_set>
                    <resource_set id="rs2">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>

                <rsc_ticket id="t2" ticket="tA">
                    <resource_set id="rs3">
                        <resource_ref id="rA"/>
                        <resource_ref id="rB"/>
                    </resource_set>
                    <resource_set id="rs4">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>

                <rsc_ticket id="t3" ticket="tB">
                    <resource_set id="rs5">
                        <resource_ref id="rA"/>
                    </resource_set>
                </rsc_ticket>
            </constraints>
        """)

        self.assertTrue(ticket.remove_with_resource_set(
            constraint_section,
            ticket_key="tA",
            resource_id="rA"
        ))

        assert_xml_equal(
            """
                <constraints>
                    <rsc_ticket id="t2" ticket="tA">
                        <resource_set id="rs3">
                            <resource_ref id="rB"/>
                        </resource_set>
                    </rsc_ticket>

                    <rsc_ticket id="t3" ticket="tB">
                        <resource_set id="rs5">
                            <resource_ref id="rA"/>
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """,
            etree.tostring(constraint_section).decode()
        )

    def test_remove_nothing_when_no_matching_found(self):
        constraint_section = etree.fromstring("""
                <constraints>
                    <rsc_ticket id="t2" ticket="tA">
                        <resource_set id="rs3">
                            <resource_ref id="rB"/>
                        </resource_set>
                    </rsc_ticket>

                    <rsc_ticket id="t3" ticket="tB">
                        <resource_set id="rs5">
                            <resource_ref id="rA"/>
                        </resource_set>
                    </rsc_ticket>
                </constraints>
        """)
        self.assertFalse(ticket.remove_with_resource_set(
            constraint_section,
            ticket_key="tA",
            resource_id="rA"
        ))



from collections import namedtuple
from enum import Enum
import threading
from typing import (
    Iterable,
    List,
    Mapping,
    Optional,
)

from lxml import etree

def run_parallel(worker, data_list):
    thread_list = []
    for args, kwargs in data_list:
        thread = threading.Thread(target=worker, args=args, kwargs=kwargs)
        thread.daemon = True
        thread_list.append(thread)
        thread.start()

    for thread in thread_list:
        thread.join()

def format_environment_error(e):
    return format_os_error(e)

def format_os_error(e):
    if e.filename:
        return "{0}: '{1}'".format(e.strerror, e.filename)
    return e.strerror

def indent(line_list: Iterable[str], indent_step: int = 2) -> List[str]:
    """
    return line list where each line of input is prefixed by N spaces

    line_list -- original lines
    indent_step -- count of spaces for line prefix
    """
    return [
        "{0}{1}".format(" " * indent_step, line) if line else line
        for line in line_list
    ]

def format_list(
    item_list: Iterable[str],
    optional_transformations: Optional[Mapping[str, str]] = None
) -> str:
    if not optional_transformations:
        optional_transformations = {}

    return ", ".join(sorted([
        "'{0}'".format(optional_transformations.get(item, item))
        for item in item_list
    ]))

def join_multilines(strings):
    return "\n".join([a.strip() for a in strings if a.strip()])

def xml_fromstring(xml):
    # If the xml contains encoding declaration such as:
    # <?xml version="1.0" encoding="UTF-8"?>
    # we get an exception in python3:
    # ValueError: Unicode strings with encoding declaration are not supported.
    # Please use bytes input or XML fragments without declaration.
    # So we encode the string to bytes.
    return etree.fromstring(
        xml.encode("utf-8"),
        #it raises on a huge xml without the flag huge_tree=True
        #see https://bugzilla.redhat.com/show_bug.cgi?id=1506864
        etree.XMLParser(huge_tree=True)
    )

class AutoNameEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):
        # pylint: disable=no-self-argument
        del start, count, last_values
        return name

class Version(namedtuple("Version", ["major", "minor", "revision"])):
    def __new__(cls, major, minor=None, revision=None):
        return super(Version, cls).__new__(cls, major, minor, revision)

    @property
    def as_full_tuple(self):
        return (
            self.major,
            self.minor if self.minor is not None else 0,
            self.revision if self.revision is not None else 0,
        )

    def normalize(self):
        return self.__class__(*self.as_full_tuple)

    def __str__(self):
        return ".".join([str(x) for x in self if x is not None])

    def __lt__(self, other):
        return self.as_full_tuple < other.as_full_tuple

    def __le__(self, other):
        return self.as_full_tuple <= other.as_full_tuple

    def __eq__(self, other):
        return self.as_full_tuple == other.as_full_tuple

    def __ne__(self, other):
        return self.as_full_tuple != other.as_full_tuple

    def __gt__(self, other):
        return self.as_full_tuple > other.as_full_tuple

    def __ge__(self, other):
        return self.as_full_tuple >= other.as_full_tuple



from functools import partial
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import(
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
)

from pcs.common import report_codes
from pcs.lib.cib.constraint import constraint
from pcs.lib.errors import ReportItemSeverity as severities

# pylint: disable=no-self-use, redundant-keyword-arg


def fixture_element(tag, _id):
    element = mock.MagicMock()
    element.tag = tag
    element.attrib = {"id": _id}
    return element

@mock.patch("pcs.lib.cib.constraint.constraint.find_parent")
@mock.patch("pcs.lib.cib.constraint.constraint.find_element_by_tag_and_id")
class FindValidResourceId(TestCase):
    def setUp(self):
        self.cib = "cib"
        self.report_processor = MockLibraryReportProcessor()
        self.find = partial(
            constraint.find_valid_resource_id,
            self.report_processor,
            self.cib,
            in_clone_allowed=False,
        )

    @staticmethod
    def fixture_error_multiinstance(parent_type, parent_id):
        return (
            severities.ERROR,
            report_codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
            {
                "resource_id": "resourceA",
                "parent_type": parent_type,
                "parent_id": parent_id,
            },
            report_codes.FORCE_CONSTRAINT_MULTIINSTANCE_RESOURCE
        )

    @staticmethod
    def fixture_warning_multiinstance(parent_type, parent_id):
        return (
            severities.WARNING,
            report_codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
            {
                "resource_id": "resourceA",
                "parent_type": parent_type,
                "parent_id": parent_id,
            },
            None
        )

    def test_return_same_id_when_resource_is_clone(self, mock_find_by_id, _):
        mock_find_by_id.return_value = fixture_element("clone", "resourceA")
        self.assertEqual("resourceA", self.find(_id="resourceA"))

    def test_return_same_id_when_resource_is_master(self, mock_find_by_id, _):
        mock_find_by_id.return_value = fixture_element("master", "resourceA")
        self.assertEqual("resourceA", self.find(_id="resourceA"))

    def test_return_same_id_when_resource_is_bundle(self, mock_find_by_id, _):
        mock_find_by_id.return_value = fixture_element("bundle", "resourceA")
        self.assertEqual("resourceA", self.find(_id="resourceA"))

    def test_return_same_id_when_resource_is_standalone_primitive(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = None
        self.assertEqual("resourceA", self.find(_id="resourceA"))

    def test_refuse_when_resource_is_in_clone(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("clone", "clone_id")
        assert_raise_library_error(
            lambda: self.find(_id="resourceA"),
            self.fixture_error_multiinstance("clone", "clone_id"),
        )

    def test_refuse_when_resource_is_in_master(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("master", "master_id")
        assert_raise_library_error(
            lambda: self.find(_id="resourceA"),
            self.fixture_error_multiinstance("clone", "master_id"),
        )

    def test_refuse_when_resource_is_in_bundle(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("bundle", "bundle_id")
        assert_raise_library_error(
            lambda: self.find(_id="resourceA"),
            self.fixture_error_multiinstance("bundle", "bundle_id"),
        )

    def test_return_resource_id_when_in_clone_allowed(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("clone", "clone_id")

        self.assertEqual(
            "resourceA",
            self.find(in_clone_allowed=True, _id="resourceA")
        )
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                self.fixture_warning_multiinstance("clone", "clone_id"),
            ]
        )

    def test_return_resource_id_when_in_master_allowed(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("master", "master_id")

        self.assertEqual(
            "resourceA",
            self.find(in_clone_allowed=True, _id="resourceA")
        )
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                self.fixture_warning_multiinstance("clone", "master_id"),
            ]
        )

    def test_return_resource_id_when_in_bundle_allowed(
         self, mock_find_by_id, mock_find_parent
    ):
        mock_find_by_id.return_value = fixture_element("primitive", "resourceA")
        mock_find_parent.return_value = fixture_element("bundle", "bundle_id")

        self.assertEqual(
            "resourceA",
            self.find(in_clone_allowed=True, _id="resourceA")
        )
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                self.fixture_warning_multiinstance("bundle", "bundle_id"),
            ]
        )

class PrepareOptionsTest(TestCase):
    def test_refuse_unknown_option(self):
        assert_raise_library_error(
            lambda: constraint.prepare_options(
                ("a", ), {"b": "c"}, mock.MagicMock(), mock.MagicMock()
            ),
            (
                severities.ERROR,
                report_codes.INVALID_OPTIONS,
                {
                    "option_names": ["b"],
                    "option_type": None,
                    "allowed": ["a", "id"],
                    "allowed_patterns": [],
                }
            ),
        )

    def test_complete_id(self):
        mock_create_id = mock.MagicMock()
        mock_create_id.return_value = "new-id"
        self.assertEqual({"id": "new-id"}, constraint.prepare_options(
            ("a",), {}, mock_create_id, mock.MagicMock()
        ))

    def test_has_no_side_efect_on_input_options(self):
        mock_create_id = mock.MagicMock()
        mock_create_id.return_value = "new-id"
        options = {"a": "b"}
        self.assertEqual(
            {"id": "new-id", "a": "b"},
            constraint.prepare_options(
                ("a",),
                options,
                mock_create_id, mock.MagicMock()
            )
        )
        self.assertEqual({"a": "b"}, options)


    def test_refuse_invalid_id(self):
        class SomeException(Exception):
            pass
        mock_validate_id = mock.MagicMock()
        mock_validate_id.side_effect = SomeException()
        self.assertRaises(
            SomeException,
            lambda: constraint.prepare_options(
                ("a", ), {"id": "invalid"}, mock.MagicMock(), mock_validate_id
            ),
        )
        mock_validate_id.assert_called_once_with("invalid")

class CreateIdTest(TestCase):
    @mock.patch(
        "pcs.lib.cib.constraint.constraint.resource_set.extract_id_set_list"
    )
    @mock.patch("pcs.lib.cib.constraint.constraint.find_unique_id")
    def test_create_id_from_resource_set_list(self, mock_find_id, mock_extract):
        mock_extract.return_value = [["A", "B"], ["C"]]
        mock_find_id.return_value = "some_id"
        self.assertEqual(
            "some_id",
            constraint.create_id("cib", "PREFIX", "resource_set_list")
        )
        mock_extract.assert_called_once_with("resource_set_list")
        mock_find_id.assert_called_once_with("cib", "pcs_PREFIX_set_A_B_set_C")

def fixture_constraint_section(return_value):
    constraint_section = mock.MagicMock()
    constraint_section.findall = mock.MagicMock()
    constraint_section.findall.return_value = return_value
    return constraint_section

@mock.patch("pcs.lib.cib.constraint.constraint.export_with_set")
class CheckIsWithoutDuplicationTest(TestCase):
    def test_raises_when_duplicate_element_found(self, export_with_set):
        export_with_set.return_value = "exported_duplicate_element"
        element = mock.MagicMock()
        element.tag = "constraint_type"

        report_processor = MockLibraryReportProcessor()
        assert_raise_library_error(
            lambda: constraint.check_is_without_duplication(
                report_processor,
                fixture_constraint_section(["duplicate_element"]), element,
                are_duplicate=lambda e1, e2: True,
                export_element=constraint.export_with_set,
            ),
            (
                severities.ERROR,
                report_codes.DUPLICATE_CONSTRAINTS_EXIST,
                {
                    'constraint_info_list': ['exported_duplicate_element'],
                    'constraint_type': 'constraint_type'
                },
                report_codes.FORCE_CONSTRAINT_DUPLICATE
            ),
        )
    def test_success_when_no_duplication_found(self, export_with_set):
        export_with_set.return_value = "exported_duplicate_element"
        element = mock.MagicMock()
        element.tag = "constraint_type"
        #no exception raised
        report_processor = MockLibraryReportProcessor()
        constraint.check_is_without_duplication(
            report_processor, fixture_constraint_section([]), element,
            are_duplicate=lambda e1, e2: True,
            export_element=constraint.export_with_set,
        )
    def test_report_when_duplication_allowed(self, export_with_set):
        export_with_set.return_value = "exported_duplicate_element"
        element = mock.MagicMock()
        element.tag = "constraint_type"

        report_processor = MockLibraryReportProcessor()
        constraint.check_is_without_duplication(
            report_processor,
            fixture_constraint_section(["duplicate_element"]), element,
            are_duplicate=lambda e1, e2: True,
            export_element=constraint.export_with_set,
            duplication_alowed=True,
        )
        assert_report_item_list_equal(
            report_processor.report_item_list,
            [
                (
                    severities.WARNING,
                    report_codes.DUPLICATE_CONSTRAINTS_EXIST,
                    {
                        'constraint_info_list': ['exported_duplicate_element'],
                        'constraint_type': 'constraint_type'
                    },
                )
            ]
        )


class CreateWithSetTest(TestCase):
    def test_put_new_constraint_to_constraint_section(self):
        constraint_section = etree.Element("constraints")
        constraint.create_with_set(
            constraint_section,
            "ticket",
            {"a": "b"},
            [{"ids": ["A", "B"], "options": {"c": "d"}}]
        )
        assert_xml_equal(etree.tostring(constraint_section).decode(), """
            <constraints>
                <ticket a="b">
                    <resource_set c="d" id="pcs_rsc_set_A_B">
                        <resource_ref id="A"/>
                        <resource_ref id="B"/>
                    </resource_set>
                </ticket>
            </constraints>
        """)

    def test_refuse_empty_resource_set_list(self):
        constraint_section = etree.Element("constraints")
        assert_raise_library_error(
            lambda: constraint.create_with_set(
                constraint_section,
                "ticket",
                {"a": "b"},
                []
            ),
            (severities.ERROR, report_codes.EMPTY_RESOURCE_SET_LIST, {})
        )



from pcs.common import report_codes
from pcs.lib import reports
from pcs.lib.errors import LibraryError
from pcs.lib.cib.node import PacemakerNode
from pcs.lib.cib.resource import primitive
from pcs.lib.resource_agent import(
    find_valid_resource_agent_by_name,
    ResourceAgentName,
)

AGENT_NAME = ResourceAgentName("ocf", "pacemaker", "remote")

def get_agent(report_processor, cmd_runner):
    return find_valid_resource_agent_by_name(
        report_processor,
        cmd_runner,
        AGENT_NAME.full_name,
    )

_IS_REMOTE_AGENT_XPATH_SNIPPET = """
    @class="{0}" and @provider="{1}" and @type="{2}"
""".format(AGENT_NAME.standard, AGENT_NAME.provider, AGENT_NAME.type)

_HAS_SERVER_XPATH_SNIPPET = """
    instance_attributes/nvpair[
        @name="server"
        and
        string-length(@value) > 0
    ]
"""

def find_node_list(tree):
    """
    Return list of remote nodes from the specified element tree

    etree.Element tree -- an element to search remote nodes in
    """
    node_list = [
        PacemakerNode(
            nvpair.getparent().getparent().attrib["id"],
            nvpair.attrib["value"]
        )
        for nvpair in tree.xpath(
            ".//primitive[{is_remote}]/{has_server}"
            .format(
                is_remote=_IS_REMOTE_AGENT_XPATH_SNIPPET,
                has_server=_HAS_SERVER_XPATH_SNIPPET,
            )
        )
    ]

    node_list.extend([
        PacemakerNode(primitive.attrib["id"], primitive.attrib["id"])
        for primitive in tree.xpath(
            ".//primitive[{is_remote} and not({has_server})]"
            .format(
                is_remote=_IS_REMOTE_AGENT_XPATH_SNIPPET,
                has_server=_HAS_SERVER_XPATH_SNIPPET,
            )
        )
    ])

    return node_list

def find_node_resources(resources_section, node_identifier):
    """
    Return list of resource elements that match to node_identifier

    etree.Element resources_section is a search element
    string node_identifier could be id of the resource or its instance attribute
        "server"
    """
    return resources_section.xpath(
        """
            .//primitive[
                {is_remote} and (
                    @id="{identifier}"
                    or
                    instance_attributes/nvpair[
                        @name="server"
                        and
                        @value="{identifier}"
                    ]
                )
            ]
        """
        .format(
            is_remote=_IS_REMOTE_AGENT_XPATH_SNIPPET,
            identifier=node_identifier
        )
    )

def get_node_name_from_resource(resource_element):
    """
    Return the node name from a remote node resource, None for other resources

    etree.Element resource_element
    """
    if not (
        resource_element.attrib.get("class", "") == AGENT_NAME.standard
        and
        resource_element.attrib.get("provider", "") == AGENT_NAME.provider
        and
        resource_element.attrib.get("type", "") == AGENT_NAME.type
    ):
        return None
    return resource_element.attrib["id"]

def _validate_server_not_used(agent, option_dict):
    if "server" in option_dict:
        return [reports.invalid_options(
            ["server"],
            sorted([
                attr["name"] for attr in agent.get_parameters()
                if attr["name"] != "server"
            ]),
            "resource",
        )]
    return []


def validate_host_not_conflicts(
    existing_nodes_addrs, node_name, instance_attributes
):
    host = instance_attributes.get("server", node_name)
    if host in existing_nodes_addrs:
        return [reports.id_already_exists(host)]
    return []

def validate_create(
    existing_nodes_names, existing_nodes_addrs, resource_agent, new_node_name,
    new_node_addr, instance_attributes
):
    """
    validate inputs for create

    list of string existing_nodes_names -- node names already in use
    list of string existing_nodes_addrs -- node addresses already in use
    ResourceAgent resource_agent -- pacemaker_remote resource agent
    string new_node_name -- the name of the future node
    string new_node_addr -- the address of the future node
    dict instance_attributes -- data for the future resource instance attributes
    """
    report_list = _validate_server_not_used(resource_agent, instance_attributes)

    addr_is_used = False
    if new_node_addr in existing_nodes_addrs:
        report_list.append(reports.id_already_exists(new_node_addr))
        addr_is_used = True

    if not addr_is_used or new_node_addr != new_node_name:
        if new_node_name in existing_nodes_names:
            report_list.append(reports.id_already_exists(new_node_name))

    return report_list

def prepare_instance_atributes(instance_attributes, host):
    enriched_instance_attributes = instance_attributes.copy()
    enriched_instance_attributes["server"] = host
    return enriched_instance_attributes

def create(
    report_processor, resource_agent, resources_section, id_provider,
    host, node_name,
    raw_operation_list=None, meta_attributes=None, instance_attributes=None,
    allow_invalid_operation=False,
    allow_invalid_instance_attributes=False,
    use_default_operations=True,
):
    # pylint: disable=too-many-arguments
    """
    Prepare all parts of remote resource and append it into the cib.

    report_processor is a tool for warning/info/error reporting
    cmd_runner is tool for launching external commands
    etree.Element resources_section is place where new element will be appended
    string node_name is name of the remote node and id of new resource as well
    list of dict raw_operation_list specifies operations of resource
    dict meta_attributes specifies meta attributes of resource
    dict instance_attributes specifies instance attributes of resource
    bool allow_invalid_operation is flag for skipping validation of operations
    bool allow_invalid_instance_attributes is flag for skipping validation of
        instance_attributes
    bool use_default_operations is flag for completion operations with default
        actions specified in resource agent
    """
    all_instance_attributes = instance_attributes.copy()
    all_instance_attributes.update({"server": host})
    try:
        return primitive.create(
            report_processor,
            resources_section,
            id_provider,
            node_name,
            resource_agent,
            raw_operation_list,
            meta_attributes,
            all_instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
    except LibraryError as e:
        for report in e.args:
            # pylint: disable=no-member
            if report.code == report_codes.INVALID_OPTIONS:
                report.info["allowed"] = [
                    value for value in report.info["allowed"]
                    if value != "server"
                ]
        raise e



from unittest import mock, TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs import settings
from pcs.common import report_codes
from pcs.common.file import RawFileError
from pcs.lib.commands import pcsd

class SynchronizeSslCertificates(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.pcsd_ssl_cert = "pcsd ssl cert"
        self.pcsd_ssl_key = "pcsd ssl key"
        self.node_names = ["rh7-1", "rh7-2"] # It matches nodes in corosync.conf
        (self.config
            .corosync_conf.load()
            .env.set_known_nodes(self.node_names)
        )

    def test_success(self):
        (self.config
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
            .http.host.send_pcsd_cert(
                cert=self.pcsd_ssl_cert,
                key=self.pcsd_ssl_key,
                node_labels=self.node_names
            )
        )

        pcsd.synchronize_ssl_certificate(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED,
                    node_name_list=self.node_names
                )
            ]
            +
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_SET_SUCCESS,
                    node=node,
                ) for node in self.node_names
            ]
        )

    def test_some_node_names_missing(self):
        nodes = ["rh7-2"]
        (self.config
            .corosync_conf.load(
                filename="corosync-some-node-names.conf",
                instead="corosync_conf.load"
            )
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
            .http.host.send_pcsd_cert(
                cert=self.pcsd_ssl_cert,
                key=self.pcsd_ssl_key,
                node_labels=nodes
            )
        )

        pcsd.synchronize_ssl_certificate(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED,
                    node_name_list=nodes
                ),
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=False,
                ),
            ]
            +
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_SET_SUCCESS,
                    node=node,
                ) for node in nodes
            ]
        )

    def test_all_node_names_missing(self):
        (self.config
            .corosync_conf.load(
                filename="corosync-no-node-names.conf",
                instead="corosync_conf.load"
            )
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: pcsd.synchronize_ssl_certificate(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=False,
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_NO_NODES_DEFINED,
                ),
            ]
        )

    def test_fail_communication(self):
        error = "an error"
        (self.config
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
            .http.host.send_pcsd_cert(
                cert=self.pcsd_ssl_cert,
                key=self.pcsd_ssl_key,
                communication_list=[
                    {
                        "label": self.node_names[0],
                        "response_code": 400,
                        "output": error,
                    }
                ] + [
                    dict(label=node) for node in self.node_names[1:]
                ]
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: pcsd.synchronize_ssl_certificate(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED,
                    node_name_list=self.node_names
                )
            ]
            +
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_SET_SUCCESS,
                    node=node,
                ) for node in self.node_names[1:]
            ]
            +
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node_names[0],
                    command="remote/set_certs",
                    reason=error
                )
            ]
        )

    def test_fail_all_nodes_unknown(self):
        (self.config
            .env.set_known_nodes([])
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: pcsd.synchronize_ssl_certificate(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=self.node_names
                ),
                fixture.error(report_codes.NONE_HOST_FOUND),
            ]
        )

    def test_fail_some_nodes_unknown(self):
        (self.config
            .env.set_known_nodes(self.node_names[1:])
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: pcsd.synchronize_ssl_certificate(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=[self.node_names[0]]
                ),
            ]
        )

    def test_some_nodes_unknown_forced(self):
        (self.config
            .env.set_known_nodes(self.node_names[1:])
            .fs.open(
                settings.pcsd_cert_location,
                mock.mock_open(read_data=self.pcsd_ssl_cert)(),
                name="fs.open.pcsd_ssl_cert"
            )
            .fs.open(
                settings.pcsd_key_location,
                mock.mock_open(read_data=self.pcsd_ssl_key)(),
                name="fs.open.pcsd_ssl_key"
            )
            .http.host.send_pcsd_cert(
                cert=self.pcsd_ssl_cert,
                key=self.pcsd_ssl_key,
                node_labels=self.node_names[1:]
            )
        )

        pcsd.synchronize_ssl_certificate(
            self.env_assist.get_env(),
            skip_offline=True
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND,
                    host_list=[self.node_names[0]]
                ),
            ]
            +
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_DISTRIBUTION_STARTED,
                    node_name_list=self.node_names[1:]
                )
            ]
            +
            [
                fixture.info(
                    report_codes.PCSD_SSL_CERT_AND_KEY_SET_SUCCESS,
                    node=node,
                ) for node in self.node_names[1:]
            ]
        )

    def test_fail_reading_files(self):
        (self.config
            .fs.open(
                settings.pcsd_cert_location,
                name="fs.open.pcsd_ssl_cert",
                side_effect=EnvironmentError(1, "error cert")
            )
            .fs.open(
                settings.pcsd_key_location,
                name="fs.open.pcsd_ssl_key",
                side_effect=EnvironmentError(1, "error key")
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: pcsd.synchronize_ssl_certificate(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.FILE_IO_ERROR,
                    file_type_code="PCSD_SSL_CERT",
                    file_path=settings.pcsd_cert_location,
                    reason="error cert",
                    operation=RawFileError.ACTION_READ,
                ),
                fixture.error(
                    report_codes.FILE_IO_ERROR,
                    file_type_code="PCSD_SSL_KEY",
                    file_path=settings.pcsd_key_location,
                    reason="error key",
                    operation=RawFileError.ACTION_READ,
                ),
            ]
        )



from collections import Counter
import re

from pcs.common import report_codes
from pcs.lib import validate
from pcs.lib.booth import constants, reports


__TICKET_NAME_RE = re.compile(r"^[\w-]+$")

def check_instance_name(name):
    """
    Check that specified booth instance name is valid

    string name -- booth instance name
    """
    report_list = []
    if "/" in name:
        # TODO drop plaintext from the report
        report_list.append(
            reports.booth_invalid_name(name, "contains illegal character '/'")
        )
    return report_list

def create(site_list, arbitrator_list):
    """
    Validate creating a minimal booth config

    iterable site_list -- list of booth sites' addresses
    iterable arbitrator_list -- list of arbitrators' addresses
    """
    report_list = []
    peer_list = site_list + arbitrator_list

    if len(site_list) < 2:
        report_list.append(reports.booth_lack_of_sites(site_list))

    if len(peer_list) % 2 == 0:
        report_list.append(reports.booth_even_peers_num(len(peer_list)))

    duplicate_addresses = {
        address for address, count in Counter(peer_list).items() if count > 1
    }
    if duplicate_addresses:
        report_list.append(
            reports.booth_address_duplication(duplicate_addresses)
        )

    return report_list

def add_ticket(
    conf_facade, ticket_name, ticket_options, allow_unknown_options=False
):
    """
    Validate adding a ticket to an existing booth config

    pcs.lib.booth.config_facade.ConfigFacade conf_facade -- a booth config
    string ticket_name -- the name of the ticket
    dict ticket_options -- ticket options
    bool allow_unknown_options -- if True, report unknown options as warnings
        instead of errors
    """
    return (
        _validate_ticket_name(ticket_name)
        +
        _validate_ticket_unique(conf_facade, ticket_name)
        +
        _validate_ticket_options(ticket_options, allow_unknown_options)
    )

def remove_ticket(conf_facade, ticket_name):
    """
    Validate removing a ticket from an existing booth config

    pcs.lib.booth.config_facade.ConfigFacade conf_facade -- a booth config
    string ticket_name -- the name of the ticket
    """
    if not conf_facade.has_ticket(ticket_name):
        return [reports.booth_ticket_does_not_exist(ticket_name)]
    return []

def _validate_ticket_name(ticket_name):
    if not __TICKET_NAME_RE.search(ticket_name):
        return [reports.booth_ticket_name_invalid(ticket_name)]
    return []

def _validate_ticket_unique(conf_facade, ticket_name):
    if conf_facade.has_ticket(ticket_name):
        return [reports.booth_ticket_duplicate(ticket_name)]
    return []

def _validate_ticket_options(options, allow_unknown_options):
    validator_list = (
        [
            validate.NamesIn(
                constants.TICKET_KEYS,
                option_type="booth ticket",
                banned_name_list=constants.GLOBAL_KEYS,
                **validate.set_warning(
                    report_codes.FORCE_OPTIONS,
                    allow_unknown_options
                )
            ),
        ]
        +
        [validate.ValueNotEmpty(option, None) for option in options]
    )
    normalized_options = validate.values_to_pairs(
        options,
        lambda key, value: value.strip()
    )
    return validate.ValidatorAll(validator_list).validate(normalized_options)



from enum import auto
from typing import (
    cast,
    Any,
    Mapping,
    Sequence,
    Union,
)

from pcs.common.interface.dto import DataTransferObject
from pcs.common.tools import AutoNameEnum


class ResourceRelationType(AutoNameEnum):
    ORDER = auto()
    ORDER_SET = auto()
    INNER_RESOURCES = auto()
    OUTER_RESOURCE = auto()


class RelationEntityDto(DataTransferObject):
    # Note: mypy doesn't understand recursive NamedTuple types, therefore this
    # class cannot inherit NamedTuple
    def __init__(
        self,
        id_: str,
        type_: Union[ResourceRelationType, str],
        members: Sequence[str],
        metadata: Mapping[str, Any],
    ):
        # pylint: disable=invalid-name
        self.id = id_
        self.type = type_
        self.members = members
        self.metadata = metadata

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, self.__class__)
            and
            self.to_dict() == cast(RelationEntityDto, other).to_dict()
        )

    def to_dict(self) -> Mapping[str, Any]:
        return dict(
            id=self.id,
            type=self.type,
            members=self.members,
            metadata=self.metadata,
        )

    @classmethod
    def from_dict(cls, payload: Mapping[str, Any]) -> "RelationEntityDto":
        obj_type = payload["type"]
        try:
            obj_type = ResourceRelationType(obj_type)
        except ValueError:
            # if obj_type is not valid ResourceRelationType, it is resource
            # type as string such as 'primitive', 'clone', ...
            pass
        return cls(
            payload["id"],
            obj_type,
            payload["members"],
            payload["metadata"],
        )


class ResourceRelationDto(DataTransferObject):
    # Note: mypy doesn't understand recursive NamedTuple types, therefore this
    # class cannot inherit NamedTuple
    def __init__(
        self,
        relation_entity: RelationEntityDto,
        members: Sequence["ResourceRelationDto"],
        is_leaf: bool,
    ):
        self.relation_entity = relation_entity
        self.members = members
        self.is_leaf = is_leaf

    def to_dict(self) -> Mapping[str, Any]:
        return dict(
            relation_entity=self.relation_entity.to_dict(),
            members=[member.to_dict() for member in self.members],
            is_leaf=self.is_leaf,
        )

    @classmethod
    def from_dict(cls, payload: Mapping[str, Any]) -> "ResourceRelationDto":
        return cls(
            RelationEntityDto.from_dict(payload["relation_entity"]),
            [
                ResourceRelationDto.from_dict(member_data)
                for member_data in payload["members"]
            ],
            payload["is_leaf"],
        )



import os.path
import site


CALL_TYPE_FS = "CALL_TYPE_FS"

_FUNC_ARGS = {
    "open": ["name", "mode", "buffering"],
    "os.path.exists": ["path"],
    "os.path.isdir": ["path"],
    "os.path.isfile": ["path"],
    "os.chmod": ["fd", "mode"],
    "os.chown": ["fd", "uid", "gid"],
    "os.listdir": ["path"],
}

def _ensure_consistent_args(func_name, call_args, call_kwargs):
    if len(call_args) > len(_FUNC_ARGS[func_name]):
        raise AssertionError(
            "{0}() too many positional arguments ({1} > {2})".format(
                func_name,
                len(call_args),
                len(_FUNC_ARGS[func_name]),
            )
        )

    param_intersection = (
        set(_FUNC_ARGS[func_name][:len(call_args)])
        .intersection(call_kwargs.keys())
    )
    if param_intersection:
        raise TypeError(
            "{0}() got multiple values for keyword argument(s) '{1}'".format(
                func_name,
                "', '".join(param_intersection)
            )
        )

def _get_all_args_as_kwargs(func_name, call_args, call_kwargs):
    _ensure_consistent_args(func_name, call_args, call_kwargs)
    kwargs = call_kwargs.copy()
    for i, arg in enumerate(call_args):
        kwargs[_FUNC_ARGS[func_name][i]] = arg
    return kwargs

class Call:
    type = CALL_TYPE_FS

    def __init__(
        self, func_name, return_value=None, side_effect=None, call_kwargs=None
    ):
        """
        callable check_stdin raises AssertionError when given stdin doesn't
            match
        """
        call_kwargs = call_kwargs if call_kwargs else {}

        #TODO side effect with return_value - mutually exclusive

        self.type = CALL_TYPE_FS
        self.func_name = func_name
        self.call_kwargs = call_kwargs
        self.return_value = return_value
        self.side_effect = side_effect

    def finish(self):
        if self.side_effect:
            if isinstance(self.side_effect, Exception):
                #pylint: disable=raising-bad-type
                raise self.side_effect
            raise AssertionError(
                "side_effect other than instance of exception not supported yet"
            )

        return self.return_value

    def __repr__(self):
        return str("<Fs '{0}' kwargs={1}>").format(
            self.func_name,
            self.call_kwargs,
        )

    def __ne__(self, other):
        return (
            self.func_name != other.func_name
            or
            self.call_kwargs != other.call_kwargs
        )

def get_fs_mock(call_queue):
    package_dir_list = site.getsitepackages()
    package_dir_list.append(os.path.realpath(
        os.path.dirname(os.path.abspath(__file__))+"/../../.."
    ))
    def get_fs_call(func_name, original_call):
        def call_fs(*args, **kwargs):
            # Standard python unittest tries to open some python code (e.g. the
            # test file for caching  when the test raises AssertionError).
            # It is before it the cleanup is called so at this moment the
            # function open is still mocked.
            # Pcs should not open file inside python package in the command so
            # attempt to open file inside pcs package is almost certainly
            # outside of library command and we will provide the original
            # function.
            if func_name == "open":
                for python_package_dir in package_dir_list:
                    if args[0].startswith(python_package_dir):
                        return original_call(*args, **kwargs)

            real_call = Call(
                func_name,
                call_kwargs=_get_all_args_as_kwargs(func_name, args, kwargs)
            )
            dummy_i, expected_call = call_queue.take(
                CALL_TYPE_FS,
                repr(real_call)
            )

            if expected_call != real_call:
                raise call_queue.error_with_context(
                    "\n  expected: '{0}'\n  but was:  '{1}'".format(
                        expected_call,
                        real_call
                    )
                )

            return expected_call.finish()
        return call_fs
    return get_fs_call

def is_fs_call_in(call_queue):
    return call_queue.has_type(CALL_TYPE_FS)



import os

from lxml import etree

from pcs_test.tools.command_env.mock_runner import (
    Call as RunnerCall,
    CheckStdinEqualXml,
)
from pcs_test.tools.fixture import complete_state_resources
from pcs_test.tools.fixture_cib import modify_cib
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str

# pylint: disable=too-many-arguments

DEFAULT_WAIT_TIMEOUT = 10
WAIT_TIMEOUT_EXPIRED_RETURNCODE = 124
AGENT_FILENAME_MAP = {
    "ocf:heartbeat:Dummy": "resource_agent_ocf_heartbeat_dummy.xml",
    "ocf:pacemaker:remote": "resource_agent_ocf_pacemaker_remote.xml",
    "ocf:heartbeat:IPaddr2": "resource_agent_ocf_heartbeat_ipaddr2.xml",
    "ocf:pacemaker:booth-site": "resource_agent_ocf_pacemaker_booth-site.xml",
}

def _fixture_state_resources_xml(
    resource_id="A", resource_agent="ocf::heartbeat:Dummy", role="Started",
    failed="false", node_name="node1"
):
    return(
        """
        <resources>
            <resource
                id="{resource_id}"
                resource_agent="{resource_agent}"
                role="{role}"
                failed="{failed}"
            >
                <node name="{node_name}" id="1" cached="false"/>
            </resource>
        </resources>
        """.format(
            resource_id=resource_id,
            resource_agent=resource_agent,
            role=role,
            failed=failed,
            node_name=node_name,
        )
    )

def _fixture_state_node_xml(
    id, name, type="member", online=True, standby=False, standby_onfail=False,
    maintenance=False, pending=False, unclean=False, shutdown=False,
    expected_up=True, is_dc=False, resources_running=0
):
    # This function uses a "clever" way of defaulting an input **dict containing
    # attributes of an xml element.
    # pylint: disable=unused-argument
    # pylint: disable=invalid-name
    # pylint: disable=redefined-builtin
    # pylint: disable=too-many-locals
    attrs = locals()
    xml_attrs = []
    for attr_name, attr_value in attrs.items():
        if attr_value is True:
            attr_value = "true"
        elif attr_value is False:
            attr_value = "false"
        xml_attrs.append('{0}="{1}"'.format(attr_name, attr_value))
    return "<node {0}/>".format(" ".join(xml_attrs))


class PcmkShortcuts():
    def __init__(self, calls):
        self.__calls = calls
        self.default_wait_timeout = DEFAULT_WAIT_TIMEOUT
        self.default_wait_error_returncode = WAIT_TIMEOUT_EXPIRED_RETURNCODE

    def fence_history_get(
        self, name="runner.pcmk.fence_history_get", node=None, stdout="",
        stderr="", returncode=0
    ):
        """
        Create call for getting plain text fencing history.

        string name -- key of the call
        string node -- a node to get a history from
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                "/usr/sbin/stonith_admin --history {0} --verbose".format(node),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def fence_history_cleanup(
        self, name="runner.pcmk.fence_history_cleanup", node=None, stdout="",
        stderr="", returncode=0
    ):
        """
        Create call for cleaning fencing history up.

        string name -- key of the call
        string node -- a node to clean a history from
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                "/usr/sbin/stonith_admin --history {0} --cleanup".format(node),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def fence_history_update(
        self, name="runner.pcmk.fence_history_update", stdout="", stderr="",
        returncode=0
    ):
        """
        Create call for updating fencing history.

        string name -- key of the call
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                "/usr/sbin/stonith_admin --history * --broadcast",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def load_state(
        self, name="runner.pcmk.load_state", filename="crm_mon.minimal.xml",
        resources=None, raw_resources=None, nodes=None, stdout="", stderr="",
        returncode=0
    ):
        """
        Create call for loading pacemaker state.

        string name -- key of the call
        string filename -- points to file with the status in the content
        string resources -- xml - resources section, will be put to state
        string nodes -- iterable of node dicts
        string stdout -- crm_mon's stdout
        string stderr -- crm_mon's stderr
        int returncode -- crm_mon's returncode
        """
        # pylint: disable=too-many-boolean-expressions
        if (
            (resources or raw_resources is not None or nodes)
            and
            (stdout or stderr or returncode)
        ):
            raise AssertionError(
                "Cannot specify resources or nodes when stdout, stderr or "
                "returncode is specified"
            )
        if resources and raw_resources is not None:
            raise AssertionError(
                "Cannot use 'resources' and 'raw_resources' together"
            )

        if (stdout or stderr or returncode):
            self.__calls.place(
                name,
                RunnerCall(
                    "crm_mon --one-shot --as-xml --inactive",
                    stdout=stdout,
                    stderr=stderr,
                    returncode=returncode
                )
            )
            return

        with open(rc(filename)) as a_file:
            state = etree.fromstring(a_file.read())

        if raw_resources is not None:
            resources = _fixture_state_resources_xml(**raw_resources)
        if resources:
            state.append(complete_state_resources(etree.fromstring(resources)))

        if nodes:
            nodes_element = state.find("./nodes")
            for node in nodes:
                nodes_element.append(
                    etree.fromstring(_fixture_state_node_xml(**node))
                )

        # set correct number of nodes and resources into the status
        resources_count = len(state.xpath(" | ".join([
            "./resources/bundle",
            "./resources/clone",
            "./resources/group",
            "./resources/resource",
        ])))
        nodes_count = len(state.findall("./nodes/node"))
        state.find("./summary/nodes_configured").set("number", str(nodes_count))
        state.find("./summary/resources_configured").set(
            "number", str(resources_count)
        )

        self.__calls.place(
            name,
            RunnerCall(
                "crm_mon --one-shot --as-xml --inactive",
                stdout=etree_to_str(state),
            )
        )

    def load_state_plaintext(
        self, name="runner.pcmk.load_state_plaintext",
        inactive=True, verbose=False, fence_history=False,
        stdout="", stderr="", returncode=0,
    ):
        """
        Create a call for loading plaintext pacemaker status

        str name -- key of the call
        bool incative -- pass --inactive flag to crm_mon
        bool verbose -- pass flags for increased verbosity to crm_mon
        bool fence_history -- pass the flag for getting fence history to crm_mon
        str stdout -- crm_mon's stdout
        str stderr -- crm_mon's stderr
        int returncode -- crm_mon's returncode
        """
        flags = ["--one-shot"]
        if inactive:
            flags.append("--inactive")
        if verbose:
            flags.extend([
                "--show-detail", "--show-node-attributes", "--failcounts"
            ])
            if fence_history:
                flags.append("--fence-history=3")
        self.__calls.place(
            name,
            RunnerCall(
                "crm_mon {}".format(" ".join(flags)),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def load_ticket_state_plaintext(
        self, name="runner.pcmk.load_ticket_state_plaintext",
        stdout="", stderr="", returncode=0,
    ):
        """
        Create a call for loading plaintext tickets status

        str name -- key of the call
        str stdout -- crm_ticket's stdout
        str stderr -- crm_ticket's stderr
        int returncode -- crm_ticket's returncode
        """
        self.__calls.place(
            name,
            RunnerCall(
                "crm_ticket --details",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def load_agent(
        self,
        name="runner.pcmk.load_agent",
        agent_name="ocf:heartbeat:Dummy",
        agent_filename=None,
        agent_is_missing=False,
        instead=None,
    ):
        """
        Create call for loading resource agent metadata.

        string name -- key of the call
        string agent_name
        string agent_filename -- points to file with the agent metadata in the
            content
        string instead -- key of call instead of which this new call is to be
            placed
        """

        if agent_filename:
            agent_metadata_filename = agent_filename
        elif agent_name in AGENT_FILENAME_MAP:
            agent_metadata_filename = AGENT_FILENAME_MAP[agent_name]
        elif not agent_is_missing:
            raise AssertionError((
                "Filename with metadata of agent '{0}' not specified.\n"
                "Please specify file with metadata for agent:\n"
                "  a) explicitly for this test:"
                " config.runner.pcmk.load_agent(agent_name='{0}',"
                " filename='FILENAME_HERE.xml')\n"
                "  b) implicitly for agent '{0}' in 'AGENT_FILENAME_MAP' in"
                " '{1}'\n"
                "Place agent metadata into '{2}FILENAME_HERE.xml'"
            ).format(agent_name, os.path.realpath(__file__), rc("")))

        if agent_is_missing:
            self.__calls.place(
                name,
                RunnerCall(
                    "crm_resource --show-metadata {0}".format(agent_name),
                    stdout="",
                    stderr=(
                        f"Agent {agent_name} not found or does not support "
                            "meta-data: Invalid argument (22)\n"
                        f"Metadata query for {agent_name} failed: Input/output "
                            "error\n"
                    ),
                    returncode=74
                ),
                instead=instead,
            )
            return

        with open(rc(agent_metadata_filename)) as a_file:
            self.__calls.place(
                name,
                RunnerCall(
                    "crm_resource --show-metadata {0}".format(agent_name),
                    stdout=a_file.read()
                ),
                instead=instead,
            )

    def load_fenced_metadata(
        self,
        name="runner.pcmk.load_fenced_metadata",
        stdout=None,
        stderr="",
        returncode=0,
        instead=None,
        before=None,
    ):
        """
        Create a call for loading fenced metadata - additional fence options

        string name -- the key of this call
        string stdout -- fenced stdout, default metadata if None
        string stderr -- fenced stderr
        int returncode -- fenced returncode
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        """
        if stdout is None:
            with open(rc("fenced_metadata.xml")) as a_file:
                stdout = a_file.read()
        self.__calls.place(
            name,
            RunnerCall(
                "/usr/libexec/pacemaker/pacemaker-fenced metadata",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode
            ),
            before=before,
            instead=instead,
        )

    def local_node_name(
        self, name="runner.pcmk.local_node_name", instead=None, before=None,
        node_name="", stdout="", stderr="", returncode=0
    ):
        """
        Create a call for crm_node --name

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string node_name -- resulting node name
        string stdout -- crm_node's stdout
        string stderr -- crm_node's stderr
        int returncode -- crm_node's returncode
        """
        if node_name and (stdout or stderr or returncode):
            raise AssertionError(
                "Cannot specify node_name when stdout, stderr or returncode is "
                "specified"
            )
        cmd = ["crm_node", "--name"]
        self.__calls.place(
            name,
            RunnerCall(
                " ".join(cmd),
                stdout=(node_name if node_name else stdout),
                stderr=stderr,
                returncode=returncode
            ),
            before=before,
            instead=instead,
        )

    def resource_cleanup(
        self,
        name="runner.pcmk.cleanup",
        instead=None,
        before=None,
        resource=None,
        node=None,
        stdout="",
        stderr="",
        returncode=0
    ):
        """
        Create a call for crm_resource --cleanup

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be cleaned
        string node -- the name of the node where resources should be cleaned
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        cmd = ["crm_resource", "--cleanup"]
        if resource:
            cmd.extend(["--resource", resource])
        if node:
            cmd.extend(["--node", node])
        self.__calls.place(
            name,
            RunnerCall(
                " ".join(cmd),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode
            ),
            before=before,
            instead=instead,
        )

    def resource_move(
        self,
        name="runner.pcmk.resource_move",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        stdout="",
        stderr="",
        returncode=0
    ):
        """
        Create a call for crm_resource --move

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be moved
        string node -- the name of a destination node
        bool master -- limit move to master role
        string lifetime -- lifetime of the created moving constraint
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--move"
        self._resource_move_ban_clear(**all_args)

    def resource_ban(
        self,
        name="runner.pcmk.resource_ban",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        stdout="",
        stderr="",
        returncode=0
    ):
        """
        Create a call for crm_resource --ban

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be banned
        string node -- the name of a destination node
        bool master -- limit ban to master role
        string lifetime -- lifetime of the created banning constraint
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--ban"
        self._resource_move_ban_clear(**all_args)

    def resource_clear(
        self,
        name="runner.pcmk.resource_clear",
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        expired=None,
        stdout="",
        stderr="",
        returncode=0
    ):
        """
        Create a call for crm_resource --clear

        string name -- the key of this call
        string instead -- the key of a call instead of which this new call is to
            be placed
        string before -- the key of a call before which this new call is to be
            placed
        string resource -- the id of a resource to be unmoved/unbanned
        string node -- the name of a destination node
        bool master -- limit clearing to master role
        bool epired -- clear only expired moves and bans
        string stdout -- crm_resource's stdout
        string stderr -- crm_resource's stderr
        int returncode -- crm_resource's returncode
        """
        # arguments are used via locals()
        # pylint: disable=unused-argument
        all_args = locals()
        del all_args["self"]
        all_args["action"] = "--clear"
        self._resource_move_ban_clear(**all_args)

    def _resource_move_ban_clear(
        self,
        name,
        action,
        instead=None,
        before=None,
        resource=None,
        node=None,
        master=None,
        lifetime=None,
        expired=None,
        stdout="",
        stderr="",
        returncode=0
    ):
        cmd = ["crm_resource", action]
        if resource:
            cmd.extend(["--resource", resource])
        if node:
            cmd.extend(["--node", node])
        if master:
            cmd.extend(["--master"])
        if lifetime:
            cmd.extend(["--lifetime", lifetime])
        if expired:
            cmd.extend(["--expired"])
        self.__calls.place(
            name,
            RunnerCall(
                " ".join(cmd),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode
            ),
            before=before,
            instead=instead,
        )

    def can_clear_expired(
        self, name="runner.pcmk.can_clear_expired", stderr="--expired"
    ):
        """
        Create call which check --expired is supported by crm_resource

        string name -- key of the call
        string stderr -- crm_resource help text
        """
        self.__calls.place(
            name,
            RunnerCall("crm_resource -?", stderr=stderr),
        )

    def wait(
        self, name="runner.pcmk.wait", stderr="", returncode=None, timeout=None
    ):
        """
        Create call for waiting to pacemaker idle

        string name -- key of the call
        string stderr -- stderr of wait command
        int returncode -- returncode of the wait command, defaults to 0 if
            stderr is empty and to 124 if stderr is not empty
        """
        if returncode is None:
            returncode = self.default_wait_error_returncode if stderr else 0

        self.__calls.place(
            name,
            RunnerCall(
                "crm_resource --wait --timeout={0}".format(
                    timeout if timeout else self.default_wait_timeout
                ),
                stderr=stderr,
                returncode=returncode,
            )
        )

    def can_wait(
        self, name="runner.pcmk.can_wait", before=None, stdout="--wait"
    ):
        """
        Create call that checks that wait for idle is supported

        string name -- key of the call
        string before -- key of call before which this new call is to be placed
        """
        self.__calls.place(
            name,
            RunnerCall("crm_resource -?", stdout=stdout),
            before=before
        )

    def verify(
        self, name="runner.pcmk.verify", cib_tempfile=None, stderr=None,
        verbose=False
    ):
        """
        Create call that checks that wait for idle is supported

        string name -- key of the call
        string before -- key of call before which this new call is to be placed
        """
        self.__calls.place(
            name,
            RunnerCall(
                "crm_verify{0} {1}".format(
                    " -V -V" if verbose else "",
                    "--xml-file {0}".format(cib_tempfile) if cib_tempfile
                        else "--live-check"
                ),
                stderr=("" if stderr is None else stderr),
                returncode=(0 if stderr is None else 55),
            ),
        )

    def remove_node(
        self, node_name,
        stderr="", returncode=0, name="runner.pcmk.remove_node",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                "crm_node --force --remove {0}".format(node_name),
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def simulate_cib(
        self, new_cib_filepath, transitions_filepath,
        cib_modifiers=None, cib_load_name="runner.cib.load",
        stdout="", stderr="", returncode=0,
        name="runner.pcmk.simulate_cib",
        **modifier_shortcuts
    ):
        """
        Create a call for simulating effects of cib changes

        string new_cib_filepath -- a temp file for storing a new cib
        string transitions_filepath -- a temp file for storing transitions
        list of callable modifiers -- every callable takes etree.Element and
            returns new etree.Element with desired modification
        string cib_load_name -- key of a call from whose stdout the cib is taken
        string stdout -- pacemaker's stdout
        string stderr -- pacemaker's stderr
        int returncode -- pacemaker's returncode
        string name -- key of the call
        dict modifier_shortcuts -- a new modifier is generated from each
            modifier shortcut.
            As key there can be keys of MODIFIER_GENERATORS.
            Value is passed into appropriate generator from MODIFIER_GENERATORS.
            For details see pcs_test.tools.fixture_cib (mainly the variable
            MODIFIER_GENERATORS - please refer it when you are adding params
            here)
        """
        cib_xml = modify_cib(
            self.__calls.get(cib_load_name).stdout,
            cib_modifiers,
            **modifier_shortcuts
        )
        cmd = [
            "crm_simulate", "--simulate",
            "--save-output", new_cib_filepath,
            "--save-graph", transitions_filepath,
            "--xml-pipe",
        ]
        self.__calls.place(
            name,
            RunnerCall(
                " ".join(cmd),
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
                check_stdin=CheckStdinEqualXml(cib_xml),
            ),
        )



from unittest import TestCase

from pcs.cli.constraint_ticket import console_report

class ConstraintPlainTest(TestCase):
    def test_prepare_report(self):
        self.assertEqual(
            "Master resourceA (id:some_id)",
            console_report.constraint_plain(
                {"options": {
                    "rsc-role": "Master",
                    "rsc": "resourceA",
                    "id": "some_id"
                }},
                with_id=True
            )
        )

    def test_prepare_report_without_role(self):
        self.assertEqual(
            "resourceA (id:some_id)",
            console_report.constraint_plain(
                {"options": {
                    "rsc": "resourceA",
                    "id": "some_id"
                }},
                with_id=True
            )
        )



from lxml import etree

def get_root(tree):
    # ElementTree has getroot, Elemet has getroottree
    return tree.getroot() if hasattr(tree, "getroot") else tree.getroottree()

def find_parent(element, tag_names):
    """
    Find parent of an element based on parent's tag name. Return the parent
    element or None if such element does not exist.

    etree element -- the element whose parent we want to find
    strings tag_names -- allowed tag names of parent we are looking for
    """
    candidate = element
    while True:
        if candidate is None or candidate.tag in tag_names:
            return candidate
        candidate = candidate.getparent()

def get_sub_element(
    element, sub_element_tag, new_id=None, new_index=None,
    append_if_missing=True
):
    """
    Returns the FIRST sub-element sub_element_tag of element. It will create new
    element if such doesn't exist yet.

    element -- parent element
    sub_element_tag -- tag of the wanted new element
    new_id -- id of the new element, None means no id will be set
    new_index -- where the new element will be added, None means at the end
    append_if_missing -- if the searched element does not exist, append it to
        the parent element
    """
    sub_element = element.find("./{0}".format(sub_element_tag))
    if sub_element is None:
        sub_element = etree.Element(sub_element_tag)
        if new_id:
            sub_element.set("id", new_id)
        if append_if_missing:
            if new_index is None:
                element.append(sub_element)
            else:
                element.insert(new_index, sub_element)
    return sub_element

def export_attributes(element):
    return  dict((key, value) for key, value in element.attrib.items())

def update_attribute_remove_empty(element, name, value):
    """
    Set an attribute's value or remove the attribute if the value is ""

    etree element -- element to be updated
    string name -- attribute name
    mixed value -- attribute value
    """
    if not value:
        if name in element.attrib:
            del element.attrib[name]
        return
    element.set(name, value)

def update_attributes_remove_empty(element, attributtes):
    """
    Set an attributes' values or remove an attribute if its new value is ""

    etree element -- element to be updated
    dict attributes -- new attributes' values
    """
    for name, value in attributtes.items():
        update_attribute_remove_empty(element, name, value)

def etree_element_attibutes_to_dict(etree_el, required_key_list):
    """
    Returns all attributes of etree_el from required_key_list in dictionary,
    where keys are attributes and values are values of attributes or None if
    it's not present.

    etree_el -- etree element from which attributes should be extracted
    required_key_list -- list of strings, attributes names which should be
        extracted
    """
    return {key: etree_el.get(key) for key in required_key_list}

def etree_to_str(tree):
    """
    Export a lxml tree to a string
    etree tree - the tree to be exported
    """
    #etree returns string in bytes: b'xml'
    #python 3 removed .encode() from byte strings
    #run(...) calls subprocess.Popen.communicate which calls encode...
    #so there is bytes to str conversion
    return etree.tostring(tree).decode()

def is_element_useful(element, attribs_important=True):
    """
    Is an element worth keeping?

    Some elements serve as a container for sub-elements. When all sub-elements
    are removed it is time to consider if such element is still meaningful.

    Some of these elements can be meaningful standalone when they contain
    attributes (e.g. "network" or "storage" in "bundle"). Some of these
    elements are not meaningful without sub-elements even if they have
    attributes (e.g. rsc_ticket - after last sub-element 'resource_set' removal
    there can be attributes but the element is pointless - more details at the
    approrpriate place of use). By default, an element is meaningful when it
    contains attributes (except id) even if it has no sub-elements. This can be
    switched by attribs_important parameter.

    lxml.etree.element element -- element to analyze
    bool attribs_important -- if True, the element is useful if it contains
        attributes even if it has no sub-elements
    """
    return len(element) or (
        attribs_important
        and
        element.attrib
        and
        element.attrib.keys() != ["id"]
    )

def append_when_useful(parent, element, attribs_important=True, index=None):
    """
    Append an element to a parent if the element is useful (see
        is_element_useful for details)

    lxml.etree.element parent -- where to append the element
    lxml.etree.element element -- the element to append
    bool attribs_important -- if True, append even if the element has no
        children if it has attributes
    int or None index -- postion to append the element, None means at the end
    """
    if element.getparent() == parent:
        return element
    if is_element_useful(element, attribs_important):
        if index is None:
            parent.append(element)
        else:
            parent.insert(index, element)
    return element

def remove_when_pointless(element, attribs_important=True):
    """
    Remove an element when it is not worth keeping (see is_element_useful for
        details).

    lxml.etree.element element -- element to remove
    bool attribs_important -- if True, do not delete the element if it contains
        attributes
    """
    if not is_element_useful(element, attribs_important):
        element.getparent().remove(element)

def reset_element(element, keep_attrs=None):
    """
    Remove all subelements and all attributes (except mentioned in keep_attrs)
    of given element.

    lxml.etree.element element -- element to reset
    list keep_attrs -- names of attributes thas should be kept
    """
    keep_attrs = keep_attrs or []
    for child in list(element):
        element.remove(child)
    for key in element.attrib.keys():
        if key not in keep_attrs:
            del element.attrib[key]



from textwrap import dedent

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall

from pcs import settings

class SystemctlShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def disable(
        self, service, name="runner.systemctl.disable",
        stdout="", stderr="", returncode=0
    ):
        self.__calls.place(
            name,
            RunnerCall(
                f"{settings.systemctl_binary} disable {service}.service",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            )
        )

    def enable(
        self, service, name="runner.systemctl.enable",
        stdout="", stderr="", returncode=0
    ):
        self.__calls.place(
            name,
            RunnerCall(
                f"{settings.systemctl_binary} enable {service}.service",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            )
        )

    def start(
        self, service, name="runner.systemctl.start",
        stdout="", stderr="", returncode=0
    ):
        self.__calls.place(
            name,
            RunnerCall(
                f"{settings.systemctl_binary} start {service}.service",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            )
        )

    def stop(
        self, service, name="runner.systemctl.stop",
        stdout="", stderr="", returncode=0
    ):
        self.__calls.place(
            name,
            RunnerCall(
                f"{settings.systemctl_binary} stop {service}.service",
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            )
        )

    def is_active(
        self, service, name="runner_systemctl.is_active", is_active=True
    ):
        args = dict(
            stdout="unknown\n",
            returncode=3,
        )
        if is_active:
            args = dict(
                stdout="active\n",
                returncode=0,
            )
        self.__calls.place(
            name,
            RunnerCall(
                "{bin_path} is-active {service}.service".format(
                    bin_path=settings.systemctl_binary,
                    service=service,
                ),
                **args
            )
        )

    def is_enabled(
        self, service, is_enabled=True,
        name="runner_systemctl.is_enabled", before=None, instead=None
    ):
        args = dict(
            stdout="disabled\n",
            returncode=1,
        )
        if is_enabled:
            args = dict(
                stdout="enabled\n",
                returncode=0,
            )
        self.__calls.place(
            name,
            RunnerCall(
                "{bin_path} is-enabled {service}.service".format(
                    bin_path=settings.systemctl_binary,
                    service=service,
                ),
                **args
            ),
            before=before,
            instead=instead
        )

    def list_unit_files(
        self, unit_file_states,
        name="runner_systemctl.list_unit_files", before=None, instead=None
    ):
        if not unit_file_states:
            output = dedent(
                """\
                UNIT FILE   STATE

                0 unit files listed.
                """
            )
        else:
            unit_len = max(len(x) for x in unit_file_states) + len(".service")
            state_len = max(len(x) for x in unit_file_states.values())
            pattern = "{unit:<{unit_len}} {state:<{state_len}}"
            lines = (
                [
                    pattern.format(
                        unit="UNIT FILE", unit_len=unit_len,
                        state="STATE", state_len=state_len
                    )
                ]
                +
                [
                    pattern.format(
                        unit="{0}.service".format(unit), unit_len=unit_len,
                        state=state, state_len=state_len
                    )
                    for unit, state in unit_file_states.items()
                ]
                +
                [
                    "",
                    "{0} unit files listed.".format(len(unit_file_states))
                ]
            )
            output = "\n".join(lines)

        self.__calls.place(
            name,
            RunnerCall(
                "{bin_path} list-unit-files --full".format(
                    bin_path=settings.systemctl_binary
                ),
                stdout=output
            ),
            before=before,
            instead=instead
        )
